#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit test module for Trait classes.

"""

from copy import copy

from anything import Something
from mock import MagicMock, patch

from evennia.objects.objects import DefaultCharacter
from evennia.utils.test_resources import BaseEvenniaTestCase, EvenniaTest

from . import traits


class _MockObj:
    def __init__(self):
        self.attributes = MagicMock()
        self.attributes.get = self.get
        self.attributes.add = self.add
        self.dbstore = {}
        self.category = "traits"

    def get(self, key, category=None):
        assert category == self.category
        return self.dbstore.get(key)

    def add(self, key, value, category=None):
        assert category == self.category
        self.dbstore[key] = value


# we want to test the base traits too
_TEST_TRAIT_CLASS_PATHS = [
    "evennia.contrib.rpg.traits.Trait",
    "evennia.contrib.rpg.traits.StaticTrait",
    "evennia.contrib.rpg.traits.CounterTrait",
    "evennia.contrib.rpg.traits.GaugeTrait",
]


class _TraitHandlerBase(BaseEvenniaTestCase):
    "Base for trait tests"

    @patch("evennia.contrib.rpg.traits.traits._TRAIT_CLASS_PATHS", new=_TEST_TRAIT_CLASS_PATHS)
    def setUp(self):
        self.obj = _MockObj()
        self.traithandler = traits.TraitHandler(self.obj)
        self.obj.traits = self.traithandler

    def _get_dbstore(self, key):
        return self.obj.dbstore["traits"][key]


class TraitHandlerTest(_TraitHandlerBase):
    """Testing for TraitHandler"""

    def setUp(self):
        super().setUp()
        self.traithandler.add("test1", name="Test1", trait_type="trait")
        self.traithandler.add(
            "test2",
            name="Test2",
            trait_type="trait",
            value=["foo", {"1": [1, 2, 3]}, 4],
        )

    def test_add_trait(self):
        self.assertEqual(
            self._get_dbstore("test1"),
            {
                "name": "Test1",
                "trait_type": "trait",
                "value": None,
            },
        )
        self.assertEqual(
            self._get_dbstore("test2"),
            {
                "name": "Test2",
                "trait_type": "trait",
                "value": ["foo", {"1": [1, 2, 3]}, 4],
            },
        )
        self.assertEqual(len(self.traithandler), 2)

    def test_cache(self):
        """
        Cache should not be set until first get
        """
        self.assertEqual(len(self.traithandler._cache), 0)
        self.traithandler.all()  # does not affect cache
        self.assertEqual(len(self.traithandler._cache), 0)
        self.traithandler.test1
        self.assertEqual(len(self.traithandler._cache), 1)
        self.traithandler.test2
        self.assertEqual(len(self.traithandler._cache), 2)

    def test_setting(self):
        "Don't allow setting stuff on traithandler"
        with self.assertRaises(traits.TraitException):
            self.traithandler.foo = "bar"
        with self.assertRaises(traits.TraitException):
            self.traithandler["foo"] = "bar"
        with self.assertRaises(traits.TraitException):
            self.traithandler.test1 = "foo"

    def test_getting(self):
        "Test we are getting data from the dbstore"
        self.assertEqual(
            self.traithandler.test1._data, {"name": "Test1", "trait_type": "trait", "value": None}
        )
        self.assertEqual(self.traithandler._cache, Something)
        self.assertEqual(
            self.traithandler.test2._data,
            {"name": "Test2", "trait_type": "trait", "value": ["foo", {"1": [1, 2, 3]}, 4]},
        )
        self.assertEqual(self.traithandler._cache, Something)
        self.assertFalse(self.traithandler.get("foo"))
        self.assertFalse(self.traithandler.bar)

    def test_all(self):
        "Test all method"
        self.assertEqual(self.traithandler.all(), ["test1", "test2"])

    def test_remove(self):
        "Test remove method"
        self.traithandler.remove("test2")
        self.assertEqual(len(self.traithandler), 1)
        self.assertTrue(bool(self.traithandler.get("test1")))  # this populates cache
        self.assertEqual(len(self.traithandler._cache), 1)
        with self.assertRaises(traits.TraitException):
            self.traithandler.remove("foo")

    def test_clear(self):
        "Test clear method"
        self.traithandler.clear()
        self.assertEqual(len(self.traithandler), 0)

    def test_trait_db_connection(self):
        "Test that updating a trait property actually updates value in db"
        trait = self.traithandler.test1
        self.assertEqual(trait.value, None)
        trait.value = 10
        self.assertEqual(trait.value, 10)
        self.assertEqual(self.obj.attributes.get("traits", category="traits")["test1"]["value"], 10)
        trait.value = 20
        self.assertEqual(trait.value, 20)
        self.assertEqual(self.obj.attributes.get("traits", category="traits")["test1"]["value"], 20)
        del trait.value
        self.assertEqual(
            self.obj.attributes.get("traits", category="traits")["test1"]["value"], None
        )


class TestTrait(_TraitHandlerBase):
    """
    Test the base Trait class
    """

    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="trait",
            value="value",
            extra_val1="xvalue1",
            extra_val2="xvalue2",
        )
        self.trait = self.traithandler.get("test1")

    def test_init(self):
        self.assertEqual(
            self.trait._data,
            {
                "name": "Test1",
                "trait_type": "trait",
                "value": "value",
                "extra_val1": "xvalue1",
                "extra_val2": "xvalue2",
            },
        )

    def test_validate_input__valid(self):
        """Test valid validation input"""
        # all data supplied, and extras
        dat = {"name": "Test", "trait_type": "trait", "value": 10, "extra_val": 1000}
        expected = copy(dat)  # we must break link or return === dat always
        self.assertEqual(expected, traits.Trait.validate_input(traits.Trait, dat))

        # don't supply value, should get default
        dat = {
            "name": "Test",
            "trait_type": "trait",
            # missing value
            "extra_val": 1000,
        }
        expected = copy(dat)
        expected["value"] = traits.Trait.default_keys["value"]
        self.assertEqual(expected, traits.Trait.validate_input(traits.Trait, dat))

        # make sure extra values are cleaned if trait accepts no extras
        dat = {
            "name": "Test",
            "trait_type": "trait",
            "value": 10,
            "extra_val1": 1000,
            "extra_val2": "xvalue",
        }
        expected = copy(dat)
        expected.pop("extra_val1")
        expected.pop("extra_val2")
        with patch.object(traits.Trait, "allow_extra_properties", False):
            self.assertEqual(expected, traits.Trait.validate_input(traits.Trait, dat))

    def test_validate_input__fail(self):
        """Test failing validation"""
        dat = {
            # missing name
            "trait_type": "trait",
            "value": 10,
            "extra_val": 1000,
        }
        with self.assertRaises(traits.TraitException):
            traits.Trait.validate_input(traits.Trait, dat)

        # make value a required key
        mock_default_keys = {"value": traits.MandatoryTraitKey}
        with patch.object(traits.Trait, "default_keys", mock_default_keys):
            dat = {
                "name": "Trait",
                "trait_type": "trait",
                # missing value, now mandatory
                "extra_val": 1000,
            }
            with self.assertRaises(traits.TraitException):
                traits.Trait.validate_input(traits.Trait, dat)

    def test_trait_getset(self):
        """Get-set-del operations on trait"""
        self.assertEqual(self.trait.name, "Test1")
        self.assertEqual(self.trait["name"], "Test1")
        self.assertEqual(self.trait.value, "value")
        self.assertEqual(self.trait["value"], "value")
        self.assertEqual(self.trait.extra_val1, "xvalue1")
        self.assertEqual(self.trait["extra_val2"], "xvalue2")

        self.trait.value = 20
        self.assertEqual(self.trait["value"], 20)
        self.trait["value"] = 20
        self.assertEqual(self.trait.value, 20)
        self.trait.extra_val1 = 100
        self.assertEqual(self.trait.extra_val1, 100)
        # additional properties
        self.trait.foo = "bar"
        self.assertEqual(self.trait.foo, "bar")

        del self.trait.foo
        with self.assertRaises(KeyError):
            self.trait["foo"]
        with self.assertRaises(AttributeError):
            self.trait.foo
        del self.trait.extra_val1
        with self.assertRaises(AttributeError):
            self.trait.extra_val1
        del self.trait.value
        # fall back to default
        self.assertTrue(self.trait.value == traits.Trait.default_keys["value"])

    def test_repr(self):
        self.assertEqual(repr(self.trait), Something)
        self.assertEqual(str(self.trait), Something)


class TestTraitStatic(_TraitHandlerBase):
    """
    Test for static Traits
    """

    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="static",
            base=1,
            mod=2,
            mult=1.0,
            extra_val1="xvalue1",
            extra_val2="xvalue2",
        )
        self.trait = self.traithandler.get("test1")

    def _get_values(self):
        return self.trait.base, self.trait.mod, self.trait.mult, self.trait.value

    def test_init(self):
        self.assertEqual(
            self._get_dbstore("test1"),
            {
                "name": "Test1",
                "trait_type": "static",
                "base": 1,
                "mod": 2,
                "mult": 1.0,
                "extra_val1": "xvalue1",
                "extra_val2": "xvalue2",
            },
        )

    def test_value(self):
        """value is (base + mod) * mult"""
        self.assertEqual(self._get_values(), (1, 2, 1.0, 3))
        self.trait.base += 4
        self.assertEqual(self._get_values(), (5, 2, 1.0, 7))
        self.trait.mod -= 1
        self.assertEqual(self._get_values(), (5, 1, 1.0, 6))
        self.trait.mult += 1.0
        self.assertEqual(self._get_values(), (5, 1, 2.0, 12))
        self.trait.mult = 0.75
        self.assertEqual(self._get_values(), (5, 1, 0.75, 4.5))

    def test_delete(self):
        """Deleting resets to default."""
        self.trait.mult = 2.0
        del self.trait.base
        self.assertEqual(self._get_values(), (0, 2, 2.0, 4))
        del self.trait.mult
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2))
        del self.trait.mod
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0))


class TestTraitCounter(_TraitHandlerBase):
    """
    Test for counter- Traits
    """

    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="counter",
            base=1,
            mod=2,
            mult=1.0,
            min=0,
            max=10,
            extra_val1="xvalue1",
            extra_val2="xvalue2",
            descs={
                0: "range0",
                2: "range1",
                5: "range2",
                7: "range3",
            },
        )
        self.trait = self.traithandler.get("test1")

    def _get_values(self):
        """Get (base, mod, mult, value, min, max)."""
        return (
            self.trait.base,
            self.trait.mod,
            self.trait.mult,
            self.trait.value,
            self.trait.min,
            self.trait.max,
        )

    def test_init(self):
        self.assertEqual(
            self._get_dbstore("test1"),
            {
                "name": "Test1",
                "trait_type": "counter",
                "base": 1,
                "mod": 2,
                "mult": 1.0,
                "min": 0,
                "max": 10,
                "extra_val1": "xvalue1",
                "extra_val2": "xvalue2",
                "descs": {
                    0: "range0",
                    2: "range1",
                    5: "range2",
                    7: "range3",
                },
                "rate": 0,
                "ratetarget": None,
                "last_update": None,
            },
        )

    def test_value(self):
        """value is (current + mod) * mult, where current defaults to base"""
        self.assertEqual(self._get_values(), (1, 2, 1.0, 3, 0, 10))
        self.trait.base += 4
        self.assertEqual(self._get_values(), (5, 2, 1.0, 7, 0, 10))
        self.trait.mod -= 1
        self.assertEqual(self._get_values(), (5, 1, 1.0, 6, 0, 10))
        self.trait.mult += 1.0
        self.assertEqual(self._get_values(), (5, 1, 2.0, 10, 0, 10))

    def test_boundaries__minmax(self):
        """Test range"""
        # should not exceed min/max values
        self.trait.base += 20
        self.assertEqual(self._get_values(), (8, 2, 1.0, 10, 0, 10))
        self.trait.base = 100
        self.assertEqual(self._get_values(), (8, 2, 1.0, 10, 0, 10))
        self.trait.base -= 40
        self.assertEqual(self._get_values(), (-2, 2, 1.0, 0, 0, 10))
        self.trait.base = -100
        self.assertEqual(self._get_values(), (-2, 2, 1.0, 0, 0, 10))

    def test_boundaries__bigmod(self):
        """add a big mod"""
        self.trait.base = 5
        self.trait.mod = 100
        self.assertEqual(self._get_values(), (5, 5, 1.0, 10, 0, 10))
        self.trait.mod = -100
        self.assertEqual(self._get_values(), (5, -5, 1.0, 0, 0, 10))

    def test_boundaries__change_boundaries(self):
        """Change boundaries after base/mod change"""
        self.trait.base = 5
        self.trait.mod = -100
        self.trait.min = -20
        self.assertEqual(self._get_values(), (5, -5, 1.0, 0, -20, 10))
        self.trait.mod -= 100
        self.assertEqual(self._get_values(), (5, -25, 1.0, -20, -20, 10))
        self.trait.mod = 100
        self.trait.max = 20
        self.assertEqual(self._get_values(), (5, 5, 1.0, 10, -20, 20))
        self.trait.mod = 100
        self.assertEqual(self._get_values(), (5, 15, 1.0, 20, -20, 20))

    def test_boundaries__disable(self):
        """Disable and re-enable boundaries"""
        self.trait.base = 5
        self.trait.mod = 100
        self.assertEqual(self._get_values(), (5, 5, 1.0, 10, 0, 10))
        del self.trait.max
        self.assertEqual(self.trait.max, None)
        del self.trait.min
        self.assertEqual(self.trait.min, None)
        self.trait.base = 100
        self.assertEqual(self._get_values(), (100, 5, 1.0, 105, None, None))
        self.trait.base = -200
        self.assertEqual(self._get_values(), (-200, 5, 1.0, -195, None, None))

        # re-activate boundaries
        self.trait.max = 15
        self.trait.min = 10  # his is blocked since base+mod is lower
        self.assertEqual(self._get_values(), (-200, 5, 1.0, -195, -195, 15))

    def test_boundaries__inverse(self):
        """Set inverse boundaries - limited by base"""
        self.trait.mod = 0
        self.assertEqual(self._get_values(), (1, 0, 1.0, 1, 0, 10))
        self.trait.min = 20  # will be set to base
        self.assertEqual(self._get_values(), (1, 0, 1.0, 1, 1, 10))
        self.trait.max = -20
        self.assertEqual(self._get_values(), (1, 0, 1.0, 1, 1, 1))

    def test_current(self):
        """Modifying current value"""
        self.trait.current = 5
        self.assertEqual(self._get_values(), (1, 2, 1.0, 7, 0, 10))
        self.trait.current = 10
        self.assertEqual(self._get_values(), (1, 2, 1.0, 10, 0, 10))
        self.trait.current = 12
        self.assertEqual(self._get_values(), (1, 2, 1.0, 10, 0, 10))
        self.trait.current = -1
        self.assertEqual(self._get_values(), (1, 2, 1.0, 2, 0, 10))
        self.trait.current -= 10
        self.assertEqual(self._get_values(), (1, 2, 1.0, 2, 0, 10))

    def test_delete(self):
        """Deleting resets to default."""
        del self.trait.base
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2, 0, 10))
        del self.trait.mod
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0, 0, 10))
        del self.trait.min
        del self.trait.max
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0, None, None))

    def test_percentage(self):
        """Test percentage calculation"""
        self.trait.base = 8
        self.trait.mod = 2
        self.trait.mult = 1.0
        self.trait.min = 0
        self.trait.max = 10
        self.assertEqual(self.trait.percent(), "100.0%")
        self.trait.current = 3
        self.assertEqual(self.trait.percent(), "50.0%")
        self.trait.current = 1
        self.assertEqual(self.trait.percent(), "30.0%")
        # have to lower this since max cannot be lowered below base+mod
        self.trait.mod = 1
        self.trait.current = 2
        self.trait.max -= 1
        self.assertEqual(self.trait.percent(), "33.3%")
        # open boundary
        del self.trait.min
        self.assertEqual(self.trait.percent(), "100.0%")

    def test_descs(self):
        """Test descriptions"""
        self.trait.min = -5
        self.trait.mod = 0
        self.assertEqual(self._get_values(), (1, 0, 1.0, 1, -5, 10))
        self.trait.current = -2
        self.assertEqual(self.trait.desc(), "range0")
        self.trait.current = 0
        self.assertEqual(self.trait.desc(), "range0")
        self.trait.current = 1
        self.assertEqual(self.trait.desc(), "range1")
        self.trait.current = 3
        self.assertEqual(self.trait.desc(), "range2")
        self.trait.current = 5
        self.assertEqual(self.trait.desc(), "range2")
        self.trait.current = 9
        self.assertEqual(self.trait.desc(), "range3")
        self.trait.current = 100
        self.assertEqual(self.trait.desc(), "range3")


class TestTraitCounterTimed(_TraitHandlerBase):
    """
    Test for trait with timer component
    """

    @patch("evennia.contrib.rpg.traits.traits.time", new=MagicMock(return_value=1000))
    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="counter",
            base=1,
            mod=2,
            mult=1.0,
            min=0,
            max=100,
            extra_val1="xvalue1",
            extra_val2="xvalue2",
            descs={
                0: "range0",
                2: "range1",
                5: "range2",
                7: "range3",
            },
            rate=1,
            ratetarget=None,
        )
        self.trait = self.traithandler.get("test1")

    def _get_timer_data(self):
        return (
            self.trait.value,
            self.trait.current,
            self.trait.rate,
            self.trait._data["last_update"],
            self.trait.ratetarget,
        )

    @patch("evennia.contrib.rpg.traits.traits.time")
    def test_timer_rate(self, mock_time):
        """Test time stepping"""
        mock_time.return_value = 1000
        self.assertEqual(self._get_timer_data(), (3, 1, 1, 1000, None))
        mock_time.return_value = 1001
        self.assertEqual(self._get_timer_data(), (4, 2, 1, 1001, None))
        mock_time.return_value = 1096
        self.assertEqual(self._get_timer_data(), (99, 97, 1, 1096, None))
        # hit maximum boundary
        mock_time.return_value = 1120
        self.assertEqual(self._get_timer_data(), (100, 98, 1, None, None))
        mock_time.return_value = 1200
        self.assertEqual(self._get_timer_data(), (100, 98, 1, None, None))
        # drop current
        self.trait.current = 50
        self.assertEqual(self._get_timer_data(), (52, 50, 1, 1200, None))
        # set a new rate
        self.trait.rate = 2
        mock_time.return_value = 1210
        self.assertEqual(self._get_timer_data(), (72, 70, 2, 1210, None))
        self.trait.rate = -10
        mock_time.return_value = 1214
        self.assertEqual(self._get_timer_data(), (32, 30, -10, 1214, None))
        mock_time.return_value = 1218
        self.assertEqual(self._get_timer_data(), (0, -2, -10, None, None))

    @patch("evennia.contrib.rpg.traits.traits.time")
    def test_timer_ratetarget(self, mock_time):
        """test ratetarget"""
        mock_time.return_value = 1000
        self.trait.ratetarget = 60
        self.assertEqual(self._get_timer_data(), (3, 1, 1, 1000, 60))
        mock_time.return_value = 1056
        self.assertEqual(self._get_timer_data(), (59, 57, 1, 1056, 60))
        mock_time.return_value = 1057
        self.assertEqual(self._get_timer_data(), (60, 58, 1, None, 60))
        mock_time.return_value = 1060
        self.assertEqual(self._get_timer_data(), (60, 58, 1, None, 60))
        self.trait.ratetarget = 70
        mock_time.return_value = 1066
        self.assertEqual(self._get_timer_data(), (66, 64, 1, 1066, 70))
        mock_time.return_value = 1070
        self.assertEqual(self._get_timer_data(), (70, 68, 1, None, 70))


class TestTraitGauge(_TraitHandlerBase):
    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="gauge",
            base=8,  # max = (base + mod) * mult
            mod=2,
            mult=1.0,
            extra_val1="xvalue1",
            extra_val2="xvalue2",
            descs={
                0: "range0",
                2: "range1",
                5: "range2",
                7: "range3",
            },
        )
        self.trait = self.traithandler.get("test1")

    def _get_values(self):
        """Get (base, mod, mult, value, min, max)."""
        return (
            self.trait.base,
            self.trait.mod,
            self.trait.mult,
            self.trait.value,
            self.trait.min,
            self.trait.max,
        )

    def test_init(self):
        self.assertEqual(
            self._get_dbstore("test1"),
            {
                "name": "Test1",
                "trait_type": "gauge",
                "base": 8,
                "mod": 2,
                "mult": 1.0,
                "min": 0,
                "extra_val1": "xvalue1",
                "extra_val2": "xvalue2",
                "descs": {
                    0: "range0",
                    2: "range1",
                    5: "range2",
                    7: "range3",
                },
                "rate": 0,
                "ratetarget": None,
                "last_update": None,
            },
        )

    def test_value(self):
        """value is current, where current defaults to base + mod"""
        # current unset - follows base + mod
        self.assertEqual(self._get_values(), (8, 2, 1.0, 10, 0, 10))
        self.trait.base += 4
        self.assertEqual(self._get_values(), (12, 2, 1.0, 14, 0, 14))
        self.trait.mod -= 1
        self.assertEqual(self._get_values(), (12, 1, 1.0, 13, 0, 13))
        self.trait.mult += 1.0
        self.assertEqual(self._get_values(), (12, 1, 2.0, 26, 0, 26))
        # set current, decouple from base + mod
        self.trait.current = 5
        self.assertEqual(self._get_values(), (12, 1, 2.0, 5, 0, 26))
        self.trait.mod += 1
        self.trait.base -= 4
        self.trait.mult -= 1.0
        self.assertEqual(self._get_values(), (8, 2, 1.0, 5, 0, 10))
        self.trait.min = -100
        self.trait.base = -20
        self.assertEqual(self._get_values(), (-20, 2, 1.0, -18, -100, -18))

    def test_boundaries__minmax(self):
        """Test range"""
        # current unset - tied to base + mod
        self.trait.base += 20
        self.assertEqual(self._get_values(), (28, 2, 1.0, 30, 0, 30))
        # set current - decouple from base + mod
        self.trait.current = 19
        self.assertEqual(self._get_values(), (28, 2, 1.0, 19, 0, 30))
        # test upper bound
        self.trait.current = 100
        self.assertEqual(self._get_values(), (28, 2, 1.0, 30, 0, 30))
        # with multiplier
        self.trait.mult = 2.0
        self.assertEqual(self._get_values(), (28, 2, 2.0, 30, 0, 60))
        self.trait.current = 100
        self.assertEqual(self._get_values(), (28, 2, 2.0, 60, 0, 60))
        # min defaults to 0
        self.trait.mult = 1.0
        self.trait.current = -10
        self.assertEqual(self._get_values(), (28, 2, 1.0, 0, 0, 30))
        self.trait.min = -20
        self.assertEqual(self._get_values(), (28, 2, 1.0, 0, -20, 30))
        self.trait.current = -10
        self.assertEqual(self._get_values(), (28, 2, 1.0, -10, -20, 30))

    def test_boundaries__bigmod(self):
        """add a big mod"""
        self.trait.base = 5
        self.trait.mod = 100
        self.assertEqual(self._get_values(), (5, 100, 1.0, 105, 0, 105))
        # restricted by min
        self.trait.mod = -100
        self.assertEqual(self._get_values(), (5, -5, 1.0, 0, 0, 0))
        self.trait.min = -200
        self.assertEqual(self._get_values(), (5, -5, 1.0, 0, -200, 0))

    def test_boundaries__change_boundaries(self):
        """Change boundaries after current change"""
        self.trait.current = 20
        self.assertEqual(self._get_values(), (8, 2, 1.0, 10, 0, 10))
        self.trait.mod = 102
        self.assertEqual(self._get_values(), (8, 102, 1.0, 10, 0, 110))
        # raising min past current value will force it upwards
        self.trait.min = 20
        self.assertEqual(self._get_values(), (8, 102, 1.0, 20, 20, 110))

    def test_boundaries__disable(self):
        """Disable and re-enable boundary"""
        self.trait.base = 5
        self.trait.min = 1
        self.assertEqual(self._get_values(), (5, 2, 1.0, 7, 1, 7))
        del self.trait.min
        self.assertEqual(self._get_values(), (5, 2, 1.0, 7, 0, 7))
        del self.trait.base
        del self.trait.mod
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0, 0, 0))
        with self.assertRaises(traits.TraitException):
            del self.trait.max

    def test_boundaries__inverse(self):
        """Try to set reversed boundaries"""
        self.trait.mod = 0
        self.trait.base = -10  # limited by min
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0, 0, 0))
        self.trait.min = -10
        self.assertEqual(self._get_values(), (0, 0, 1.0, 0, -10, 0))
        self.trait.base = -10
        self.assertEqual(self._get_values(), (-10, 0, 1.0, -10, -10, -10))
        self.min = 0  # limited by base + mod
        self.assertEqual(self._get_values(), (-10, 0, 1.0, -10, -10, -10))

    def test_current(self):
        """Modifying current value"""
        self.trait.base = 10
        self.trait.current = 5
        self.assertEqual(self._get_values(), (10, 2, 1.0, 5, 0, 12))
        self.trait.current = 10
        self.assertEqual(self._get_values(), (10, 2, 1.0, 10, 0, 12))
        self.trait.current = 12
        self.assertEqual(self._get_values(), (10, 2, 1.0, 12, 0, 12))
        self.trait.current = 0
        self.assertEqual(self._get_values(), (10, 2, 1.0, 0, 0, 12))
        self.trait.current = -1
        self.assertEqual(self._get_values(), (10, 2, 1.0, 0, 0, 12))

    def test_delete(self):
        """Deleting resets to default."""
        del self.trait.mod
        self.assertEqual(self._get_values(), (8, 0, 1.0, 8, 0, 8))
        self.trait.mod = 2
        del self.trait.base
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2, 0, 2))
        del self.trait.min
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2, 0, 2))
        self.trait.min = -10
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2, -10, 2))
        del self.trait.min
        self.assertEqual(self._get_values(), (0, 2, 1.0, 2, 0, 2))

    def test_percentage(self):
        """Test percentage calculation"""
        self.assertEqual(self.trait.percent(), "100.0%")
        self.trait.current = 5
        self.assertEqual(self.trait.percent(), "50.0%")
        self.trait.current = 3
        self.assertEqual(self.trait.percent(), "30.0%")
        self.trait.mod -= 1
        self.assertEqual(self.trait.percent(), "33.3%")

    def test_descs(self):
        """Test descriptions"""
        self.trait.min = -5
        self.assertEqual(self._get_values(), (8, 2, 1.0, 10, -5, 10))
        self.trait.current = -2
        self.assertEqual(self.trait.desc(), "range0")
        self.trait.current = 0
        self.assertEqual(self.trait.desc(), "range0")
        self.trait.current = 1
        self.assertEqual(self.trait.desc(), "range1")
        self.trait.current = 3
        self.assertEqual(self.trait.desc(), "range2")
        self.trait.current = 5
        self.assertEqual(self.trait.desc(), "range2")
        self.trait.current = 9
        self.assertEqual(self.trait.desc(), "range3")
        self.trait.current = 100
        self.assertEqual(self.trait.desc(), "range3")


class TestTraitGaugeTimed(_TraitHandlerBase):
    """
    Test for trait with timer component
    """

    @patch("evennia.contrib.rpg.traits.traits.time", new=MagicMock(return_value=1000))
    def setUp(self):
        super().setUp()
        self.traithandler.add(
            "test1",
            name="Test1",
            trait_type="gauge",
            base=98,
            mod=2,
            min=0,
            extra_val1="xvalue1",
            extra_val2="xvalue2",
            descs={
                0: "range0",
                2: "range1",
                5: "range2",
                7: "range3",
            },
            rate=1,
            ratetarget=None,
        )
        self.trait = self.traithandler.get("test1")

    def _get_timer_data(self):
        return (
            self.trait.value,
            self.trait.current,
            self.trait.rate,
            self.trait._data["last_update"],
            self.trait.ratetarget,
        )

    @patch("evennia.contrib.rpg.traits.traits.time")
    def test_timer_rate(self, mock_time):
        """Test time stepping"""
        mock_time.return_value = 1000
        self.trait.current = 1
        self.assertEqual(self._get_timer_data(), (1, 1, 1, 1000, None))
        mock_time.return_value = 1001
        self.assertEqual(self._get_timer_data(), (2, 2, 1, 1001, None))
        mock_time.return_value = 1096
        self.assertEqual(self._get_timer_data(), (97, 97, 1, 1096, None))
        # hit maximum boundary
        mock_time.return_value = 1120
        self.assertEqual(self._get_timer_data(), (100, 100, 1, None, None))
        mock_time.return_value = 1200
        self.assertEqual(self._get_timer_data(), (100, 100, 1, None, None))
        # drop current
        self.trait.current = 50
        self.assertEqual(self._get_timer_data(), (50, 50, 1, 1200, None))
        # set a new rate
        self.trait.rate = 2
        mock_time.return_value = 1210
        self.assertEqual(self._get_timer_data(), (70, 70, 2, 1210, None))
        self.trait.rate = -10
        mock_time.return_value = 1214
        self.assertEqual(self._get_timer_data(), (30, 30, -10, 1214, None))
        mock_time.return_value = 1218
        self.assertEqual(self._get_timer_data(), (0, 0, -10, None, None))

    @patch("evennia.contrib.rpg.traits.traits.time")
    def test_timer_ratetarget(self, mock_time):
        """test ratetarget"""
        mock_time.return_value = 1000
        self.trait.current = 1
        self.trait.ratetarget = 60
        self.assertEqual(self._get_timer_data(), (1, 1, 1, 1000, 60))
        mock_time.return_value = 1056
        self.assertEqual(self._get_timer_data(), (57, 57, 1, 1056, 60))
        mock_time.return_value = 1059
        self.assertEqual(self._get_timer_data(), (60, 60, 1, None, 60))
        mock_time.return_value = 1060
        self.assertEqual(self._get_timer_data(), (60, 60, 1, None, 60))
        self.trait.ratetarget = 70
        mock_time.return_value = 1066
        self.assertEqual(self._get_timer_data(), (66, 66, 1, 1066, 70))
        mock_time.return_value = 1070
        self.assertEqual(self._get_timer_data(), (70, 70, 1, None, 70))


class TestNumericTraitOperators(BaseEvenniaTestCase):
    """Test case for numeric magic method implementations."""

    def setUp(self):
        # direct instantiation for testing only; use TraitHandler in production
        self.st = traits.Trait(
            {
                "name": "Strength",
                "trait_type": "trait",
                "value": 8,
            }
        )
        self.at = traits.Trait(
            {
                "name": "Attack",
                "trait_type": "trait",
                "value": 4,
            }
        )

    def tearDown(self):
        self.st, self.at = None, None

    def test_pos_shortcut(self):
        """overridden unary + operator returns `value` property"""
        self.assertIn(type(+self.st), (float, int))
        self.assertEqual(+self.st, self.st.value)
        self.assertEqual(+self.st, 8)

    def test_add_traits(self):
        """test addition of `Trait` objects"""
        # two Trait objects
        self.assertEqual(self.st + self.at, 12)
        # Trait and numeric
        self.assertEqual(self.st + 1, 9)
        self.assertEqual(1 + self.st, 9)

    def test_sub_traits(self):
        """test subtraction of `Trait` objects"""
        # two Trait objects
        self.assertEqual(self.st - self.at, 4)
        # Trait and numeric
        self.assertEqual(self.st - 1, 7)
        self.assertEqual(10 - self.st, 2)

    def test_mul_traits(self):
        """test multiplication of `Trait` objects"""
        # between two Traits
        self.assertEqual(self.st * self.at, 32)
        # between Trait and numeric
        self.assertEqual(self.at * 4, 16)
        self.assertEqual(4 * self.at, 16)

    def test_floordiv(self):
        """test floor division of `Trait` objects"""
        # between two Traits
        self.assertEqual(self.st // self.at, 2)
        # between Trait and numeric
        self.assertEqual(self.st // 2, 4)
        self.assertEqual(18 // self.st, 2)

    def test_comparisons_traits(self):
        """test equality comparison between `Trait` objects"""
        self.assertNotEqual(self.st, self.at)
        self.assertLess(self.at, self.st)
        self.assertLessEqual(self.at, self.st)
        self.assertGreater(self.st, self.at)
        self.assertGreaterEqual(self.st, self.at)

    def test_comparisons_numeric(self):
        """equality comparisons between `Trait` and numeric"""
        self.assertEqual(self.st, 8)
        self.assertEqual(8, self.st)
        self.assertNotEqual(self.st, 0)
        self.assertNotEqual(0, self.st)
        self.assertLess(self.st, 10)
        self.assertLess(0, self.st)
        self.assertLessEqual(self.st, 8)
        self.assertLessEqual(8, self.st)
        self.assertLessEqual(self.st, 10)
        self.assertLessEqual(0, self.st)
        self.assertGreater(self.st, 0)
        self.assertGreater(10, self.st)
        self.assertGreaterEqual(self.st, 8)
        self.assertGreaterEqual(8, self.st)
        self.assertGreaterEqual(self.st, 0)
        self.assertGreaterEqual(10, self.st)


class DummyCharacter(_MockObj):
    strength = traits.TraitProperty("Strength", trait_type="static", base=10, mod=2)
    hunting = traits.TraitProperty("Hunting skill", trait_type="counter", base=10, mod=1, max=100)
    health = traits.TraitProperty("Health value", trait_type="gauge", base=100)


class TestTraitFields(BaseEvenniaTestCase):
    """
    Test the TraitField class.

    """

    @patch("evennia.contrib.rpg.traits.traits._TRAIT_CLASS_PATHS", new=_TEST_TRAIT_CLASS_PATHS)
    def test_traitfields(self):
        obj = DummyCharacter()
        obj2 = DummyCharacter()

        self.assertEqual(12, obj.strength.value)
        self.assertEqual(11, obj.hunting.value)
        self.assertEqual(100, obj.health.value)

        obj.strength.base += 5
        self.assertEqual(17, obj.strength.value)

        obj.strength.berserk = True
        self.assertEqual(obj.strength.berserk, True)

        self.assertEqual(100, obj.traits.health)
        self.assertEqual(None, obj.traits.hp)

        # the traithandler still works
        obj.traits.health.current -= 1
        self.assertEqual(99, obj.health.value)

        # making sure Descriptors are separate
        self.assertEqual(12, obj2.strength.value)
        self.assertEqual(17, obj.strength.value)

        obj2.strength.base += 1
        obj.strength.base += 3

        self.assertEqual(13, obj2.strength.value)
        self.assertEqual(20, obj.strength.value)


class TraitContribTestingChar(DefaultCharacter):
    HP = traits.TraitProperty("health", trait_type="trait", value=5)


class TraitPropertyTestCase(EvenniaTest):
    """
    Test atomic updating.

    """

    character_typeclass = TraitContribTestingChar

    def test_round1(self):
        self.char1.HP.value = 1

    def test_round2(self):
        self.char1.HP.value = 2



"""
Slow Exit typeclass

Contribution - Griatch 2014


This is an example of an Exit-type that delays its traversal. This
simulates slow movement, common in many different types of games. The
contrib also contains two commands, `CmdSetSpeed` and CmdStop for changing
the movement speed and abort an ongoing traversal, respectively.

## Installation:

To try out an exit of this type, you could connect two existing rooms
using something like this:

@open north:contrib.grid.slow_exit.SlowExit = <destination>

To make this your new default exit, modify `mygame/typeclasses/exits.py`
to import this module and change the default `Exit` class to inherit
from `SlowExit` instead.

```
# in mygame/typeclasses/exits.py

from evennia.contrib.grid.slowexit import SlowExit

class Exit(SlowExit):
    # ...

```

To get the ability to change your speed and abort your movement, import

```python
# in mygame/commands/default_cmdsets.py

from evennia.contrib.grid import slow_exit  <---

class CharacterCmdSet(default_cmds.CharacterCmdSet):
    # ...
    def at_cmdset_creation(self):
        # ...
        self.add(slow_exit.SlowDoorCmdSet)  <---

```

## Notes:

This implementation is efficient but not persistent; so incomplete
movement will be lost in a server reload. This is acceptable for most
game types - to simulate longer travel times (more than the couple of
seconds assumed here), a more persistent variant using Scripts or the
TickerHandler might be better.

"""

from evennia.commands.cmdset import CmdSet
from evennia.commands.command import Command
from evennia.objects.objects import DefaultExit
from evennia.utils import utils

MOVE_DELAY = {"stroll": 6, "walk": 4, "run": 2, "sprint": 1}


class SlowExit(DefaultExit):
    """
    This overloads the way moving happens.
    """

    def at_traverse(self, traversing_object, target_location):
        """
        Implements the actual traversal, using utils.delay to delay the move_to.
        """

        # if the traverser has an Attribute move_speed, use that,
        # otherwise default to "walk" speed
        move_speed = traversing_object.db.move_speed or "walk"
        move_delay = MOVE_DELAY.get(move_speed, 4)

        def move_callback():
            "This callback will be called by utils.delay after move_delay seconds."
            source_location = traversing_object.location
            if traversing_object.move_to(target_location, move_type="traverse"):
                self.at_post_traverse(traversing_object, source_location)
            else:
                if self.db.err_traverse:
                    # if exit has a better error message, let's use it.
                    self.caller.msg(self.db.err_traverse)
                else:
                    # No shorthand error message. Call hook.
                    self.at_failed_traverse(traversing_object)

        traversing_object.msg("You start moving %s at a %s." % (self.key, move_speed))
        # create a delayed movement
        t = utils.delay(move_delay, move_callback)
        # we store the deferred on the character, this will allow us
        # to abort the movement. We must use an ndb here since
        # deferreds cannot be pickled.
        traversing_object.ndb.currently_moving = t


#
# set speed - command
#

SPEED_DESCS = {"stroll": "strolling", "walk": "walking", "run": "running", "sprint": "sprinting"}


class CmdSetSpeed(Command):
    """
    set your movement speed

    Usage:
      setspeed stroll|walk|run|sprint

    This will set your movement speed, determining how long time
    it takes to traverse exits. If no speed is set, 'walk' speed
    is assumed.
    """

    key = "setspeed"

    def func(self):
        """
        Simply sets an Attribute used by the SlowExit above.
        """
        speed = self.args.lower().strip()
        if speed not in SPEED_DESCS:
            self.caller.msg("Usage: setspeed stroll||walk||run||sprint")
        elif self.caller.db.move_speed == speed:
            self.caller.msg("You are already %s." % SPEED_DESCS[speed])
        else:
            self.caller.db.move_speed = speed
            self.caller.msg("You are now %s." % SPEED_DESCS[speed])


#
# stop moving - command
#


class CmdStop(Command):
    """
    stop moving

    Usage:
      stop

    Stops the current movement, if any.
    """

    key = "stop"

    def func(self):
        """
        This is a very simple command, using the
        stored deferred from the exit traversal above.
        """
        currently_moving = self.caller.ndb.currently_moving
        if currently_moving and not currently_moving.called:
            currently_moving.cancel()
            self.caller.msg("You stop moving.")
            for observer in self.caller.location.contents_get(self.caller):
                observer.msg("%s stops." % self.caller.get_display_name(observer))
        else:
            self.caller.msg("You are not moving.")


class SlowExitCmdSet(CmdSet):
    def at_cmdset_creation(self):
        self.add(CmdSetSpeed())
        self.add(CmdStop())



from django.conf import settings
from django.test import Client, override_settings
from django.urls import reverse
from django.utils.text import slugify

from evennia.help import filehelp
from evennia.utils import class_from_module
from evennia.utils.create import create_help_entry
from evennia.utils.test_resources import BaseEvenniaTest

_FILE_HELP_ENTRIES = None


class EvenniaWebTest(BaseEvenniaTest):
    # Use the same classes the views are expecting
    account_typeclass = settings.BASE_ACCOUNT_TYPECLASS
    object_typeclass = settings.BASE_OBJECT_TYPECLASS
    character_typeclass = settings.BASE_CHARACTER_TYPECLASS
    exit_typeclass = settings.BASE_EXIT_TYPECLASS
    room_typeclass = settings.BASE_ROOM_TYPECLASS
    script_typeclass = settings.BASE_SCRIPT_TYPECLASS
    channel_typeclass = settings.BASE_CHANNEL_TYPECLASS

    # Default named url
    url_name = "index"

    # Response to expect for unauthenticated requests
    unauthenticated_response = 200

    # Response to expect for authenticated requests
    authenticated_response = 200

    def setUp(self):
        super().setUp()

        # Add chars to account rosters
        self.account.characters.add(self.char1)
        self.account2.characters.add(self.char2)

        for account in (self.account, self.account2):
            # Demote accounts to Player permissions
            account.permissions.add("Player")
            account.permissions.remove("Developer")

            # Grant permissions to chars
            for char in account.characters:
                char.locks.add("edit:id(%s) or perm(Admin)" % account.pk)
                char.locks.add("delete:id(%s) or perm(Admin)" % account.pk)
                char.locks.add("view:all()")

    def test_valid_chars(self):
        "Make sure account has playable characters"
        self.assertTrue(self.char1 in self.account.characters)
        self.assertTrue(self.char2 in self.account2.characters)

    def get_kwargs(self):
        return {}

    def test_get(self):
        # Try accessing page while not logged in
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()))
        self.assertEqual(response.status_code, self.unauthenticated_response)

    def login(self):
        return self.client.login(username="TestAccount", password="testpassword")

    def test_get_authenticated(self):
        logged_in = self.login()
        self.assertTrue(logged_in, "Account failed to log in!")

        # Try accessing page while logged in
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)

        self.assertEqual(response.status_code, self.authenticated_response)


# ------------------------------------------------------------------------------


class AdminTest(EvenniaWebTest):
    url_name = "django_admin"
    unauthenticated_response = 302


class IndexTest(EvenniaWebTest):
    url_name = "index"


class RegisterTest(EvenniaWebTest):
    url_name = "register"


class LoginTest(EvenniaWebTest):
    url_name = "login"


class LogoutTest(EvenniaWebTest):
    url_name = "logout"


class PasswordResetTest(EvenniaWebTest):
    url_name = "password_change"
    unauthenticated_response = 302


class WebclientTest(EvenniaWebTest):
    url_name = "webclient:index"

    @override_settings(WEBCLIENT_ENABLED=True)
    def test_get(self):
        self.authenticated_response = 200
        self.unauthenticated_response = 200
        super().test_get()

    @override_settings(WEBCLIENT_ENABLED=False)
    def test_get_disabled(self):
        self.authenticated_response = 404
        self.unauthenticated_response = 404
        super().test_get()


class ChannelListTest(EvenniaWebTest):
    url_name = "channels"


class ChannelDetailTest(EvenniaWebTest):
    url_name = "channel-detail"

    def setUp(self):
        super().setUp()

        klass = class_from_module(
            self.channel_typeclass, fallback=settings.FALLBACK_CHANNEL_TYPECLASS
        )

        # Create a channel
        klass.create("demo")

    def get_kwargs(self):
        return {"slug": slugify("demo")}


class HelpListTest(EvenniaWebTest):
    url_name = "help"


HELP_ENTRY_DICTS = [
    {"key": "unit test file entry", "category": "General", "text": "cache test file entry text"}
]


class HelpDetailTest(EvenniaWebTest):
    url_name = "help-entry-detail"

    def setUp(self):
        super().setUp()

        # create a db help entry
        create_help_entry("unit test db entry", "unit test db entry text", category="General")

    def get_kwargs(self):
        return {"category": slugify("general"), "topic": slugify("unit test db entry")}

    def test_view(self):
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.context["entry_text"], "unit test db entry text")

    def test_object_cache(self):
        # clear file help entries, use local HELP_ENTRY_DICTS to recreate new entries
        global _FILE_HELP_ENTRIES
        if _FILE_HELP_ENTRIES is None:
            from evennia.help.filehelp import FILE_HELP_ENTRIES as _FILE_HELP_ENTRIES
        help_module = "evennia.web.website.tests"
        self.file_help_store = _FILE_HELP_ENTRIES.__init__(help_file_modules=[help_module])

        # request access to an entry
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.context["entry_text"], "unit test db entry text")
        # request a second entry, verifing the cached object is not provided on a new topic request
        entry_two_args = {"category": slugify("general"), "topic": slugify("unit test file entry")}
        response = self.client.get(reverse(self.url_name, kwargs=entry_two_args), follow=True)
        self.assertEqual(response.context["entry_text"], "cache test file entry text")


class HelpLockedDetailTest(EvenniaWebTest):
    url_name = "help-entry-detail"

    def setUp(self):
        super().setUp()

        # create a db entry with a lock
        self.db_help_entry = create_help_entry(
            "unit test locked topic",
            "unit test locked entrytext",
            category="General",
            locks="read:perm(Developer)",
        )

    def get_kwargs(self):
        return {"category": slugify("general"), "topic": slugify("unit test locked topic")}

    def test_locked_entry(self):
        # request access to an entry for permission the account does not have
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.context["entry_text"], "Failed to find entry.")

    def test_lock_with_perm(self):
        # log TestAccount in, grant permission required, read the entry
        self.login()
        self.account.permissions.add("Developer")
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.context["entry_text"], "unit test locked entrytext")


class CharacterCreateView(EvenniaWebTest):
    url_name = "character-create"
    unauthenticated_response = 302

    @override_settings(MAX_NR_CHARACTERS=1)
    def test_valid_access_multisession_0(self):
        "Account1 with no characters should be able to create a new one"
        self.account.characters.remove(self.char1)

        # Login account
        self.login()

        # Post data for a new character
        data = {"db_key": "gannon", "desc": "Some dude."}

        response = self.client.post(reverse(self.url_name), data=data, follow=True)
        self.assertEqual(response.status_code, 200)

        # Make sure the character was actually created
        self.assertTrue(
            len(self.account.characters) == 1,
            "Account only has the following characters attributed to it: %s"
            % self.account.characters,
        )

    @override_settings(MAX_NR_CHARACTERS=5)
    def test_valid_access_multisession_2(self):
        "Account1 should be able to create multiple new characters"
        # Login account
        self.login()

        # Post data for a new character
        data = {"db_key": "gannon", "desc": "Some dude."}

        response = self.client.post(reverse(self.url_name), data=data, follow=True)
        self.assertEqual(response.status_code, 200)

        # Make sure the character was actually created
        self.assertTrue(
            len(self.account.characters) > 1,
            "Account only has the following characters attributed to it: %s"
            % self.account.characters,
        )


class CharacterPuppetView(EvenniaWebTest):
    url_name = "character-puppet"
    unauthenticated_response = 302

    def get_kwargs(self):
        return {"pk": self.char1.pk, "slug": slugify(self.char1.name)}

    def test_invalid_access(self):
        "Account1 should not be able to puppet Account2:Char2"
        # Login account
        self.login()

        # Try to access puppet page for char2
        kwargs = {"pk": self.char2.pk, "slug": slugify(self.char2.name)}
        response = self.client.get(reverse(self.url_name, kwargs=kwargs), follow=True)
        self.assertTrue(
            response.status_code >= 400,
            "Invalid access should return a 4xx code-- either obj not found or permission denied!"
            " (Returned %s)" % response.status_code,
        )


class CharacterListView(EvenniaWebTest):
    url_name = "characters"
    unauthenticated_response = 302


class CharacterManageView(EvenniaWebTest):
    url_name = "character-manage"
    unauthenticated_response = 302


class CharacterUpdateView(EvenniaWebTest):
    url_name = "character-update"
    unauthenticated_response = 302

    def get_kwargs(self):
        return {"pk": self.char1.pk, "slug": slugify(self.char1.name)}

    def test_valid_access(self):
        "Account1 should be able to update Account1:Char1"
        # Login account
        self.login()

        # Try to access update page for char1
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.status_code, 200)

        # Try to update char1 desc
        data = {"db_key": self.char1.db_key, "desc": "Just a regular type of dude."}
        response = self.client.post(
            reverse(self.url_name, kwargs=self.get_kwargs()), data=data, follow=True
        )
        self.assertEqual(response.status_code, 200)

        # Make sure the change was made successfully
        self.assertEqual(self.char1.db.desc, data["desc"])

    def test_invalid_access(self):
        "Account1 should not be able to update Account2:Char2"
        # Login account
        self.login()

        # Try to access update page for char2
        kwargs = {"pk": self.char2.pk, "slug": slugify(self.char2.name)}
        response = self.client.get(reverse(self.url_name, kwargs=kwargs), follow=True)
        self.assertEqual(response.status_code, 403)


class CharacterDeleteView(EvenniaWebTest):
    url_name = "character-delete"
    unauthenticated_response = 302

    def get_kwargs(self):
        return {"pk": self.char1.pk, "slug": slugify(self.char1.name)}

    def test_valid_access(self):
        "Account1 should be able to delete Account1:Char1"
        # Login account
        self.login()

        # Try to access delete page for char1
        response = self.client.get(reverse(self.url_name, kwargs=self.get_kwargs()), follow=True)
        self.assertEqual(response.status_code, 200)

        # Proceed with deleting it
        data = {"value": "yes"}
        response = self.client.post(
            reverse(self.url_name, kwargs=self.get_kwargs()), data=data, follow=True
        )
        self.assertEqual(response.status_code, 200)

        # Make sure it deleted
        self.assertFalse(
            self.char1 in self.account.characters,
            "Char1 is still in Account playable characters list.",
        )

    def test_invalid_access(self):
        "Account1 should not be able to delete Account2:Char2"
        # Login account
        self.login()

        # Try to access delete page for char2
        kwargs = {"pk": self.char2.pk, "slug": slugify(self.char2.name)}
        response = self.client.get(reverse(self.url_name, kwargs=kwargs), follow=True)
        self.assertEqual(response.status_code, 403)



"""
Room class and mechanics for the Evscaperoom.

This is a special room class that not only depicts the evscaperoom itself, it
also acts as a central store for the room state, score etc. When deleting this,
that particular escaperoom challenge should be gone.

"""

from evennia import DefaultCharacter, DefaultObject, DefaultRoom, logger, utils
from evennia.locks.lockhandler import check_lockstring
from evennia.utils.ansi import strip_ansi
from evennia.utils.utils import lazy_property, list_to_string

from .commands import CmdSetEvScapeRoom
from .objects import EvscaperoomObject
from .state import StateHandler


class EvscapeRoom(EvscaperoomObject, DefaultRoom):
    """
    The room to escape from.

    """

    def at_object_creation(self):
        """
        Called once, when the room is first created.

        """
        super().at_object_creation()

        # starting state
        self.db.state = None  # name
        self.db.prev_state = None

        # this is used for tagging of all objects belonging to this
        # particular room instance, so they can be cleaned up later
        # this is accessed through the .tagcategory getter.
        self.db.tagcategory = "evscaperoom_{}".format(self.key)

        # room progress statistics
        self.db.stats = {
            "progress": 0,  # in percent
            "score": {},  # reason: score
            "max_score": 100,
            "hints_used": 0,  # total across all states
            "hints_total": 41,
            "total_achievements": 14,
        }

        self.cmdset.add(CmdSetEvScapeRoom, persistent=True)

        self.log("Room created and log started.")

    @lazy_property
    def statehandler(self):
        return StateHandler(self)

    @property
    def state(self):
        return self.statehandler.current_state

    def log(self, message, caller=None):
        """
        Log to a file specificially for this room.
        """
        caller = f"[caller.key]: " if caller else ""

        logger.log_file(
            strip_ansi(f"{caller}{message.strip()}"), filename=self.tagcategory + ".log"
        )

    def score(self, new_score, reason):
        """
        We don't score individually but for everyone in room together.
        You can only be scored for a given reason once."""
        if reason not in self.db.stats["score"]:
            self.log(f"score: {reason} ({new_score}pts)")
            self.db.stats["score"][reason] = new_score

    def progress(self, new_progress):
        "Progress is what we set it to be (0-100%)"
        self.log(f"progress: {new_progress}%")
        self.db.stats["progress"] = new_progress

    def achievement(self, caller, achievement, subtext=""):
        """
        Give the caller a personal achievment. You will only
        ever get one of the same type

        Args:
            caller (Object): The receiver of the achievement.
            achievement (str): The title/name of the achievement.
            subtext (str, optional): Eventual subtext/explanation
                of the achievement.
        """
        achievements = caller.attributes.get("achievements", category=self.tagcategory)
        if not achievements:
            achievements = {}
        if achievement not in achievements:
            self.log(f"achievement: {caller} earned '{achievement}' - {subtext}")
            achievements[achievement] = subtext
            caller.attributes.add("achievements", achievements, category=self.tagcategory)

    def get_all_characters(self):
        """
        Get the player characters in the room.

        Returns:
            chars (Queryset): The characters.

        """
        return DefaultCharacter.objects.filter_family(db_location=self)

    def set_flag(self, flagname):
        self.db.flags[flagname] = True

    def unset_flag(self, flagname):
        if flagname in self.db.flags:
            del self.db.flags[flagname]

    def check_flag(self, flagname):
        return self.db.flags.get(flagname, False)

    def check_perm(self, caller, permission):
        return check_lockstring(caller, f"dummy:perm({permission})")

    def tag_character(self, character, tag, category=None):
        """
        Tag a given character in this room.

        Args:
            character (Character): Player character to tag.
            tag (str): Tag to set.
            category (str, optional): Tag-category. If unset, use room's
                tagcategory.

        """
        category = category if category else self.db.tagcategory
        character.tags.add(tag, category=category)

    def tag_all_characters(self, tag, category=None):
        """
        Set a given tag on all players in the room.

        Args:
            room (EvscapeRoom): The room to escape from.
            tag (str): The tag to set.
            category (str, optional): If unset, will use the room's tagcategory.

        """
        category = category if category else self.tagcategory

        for char in self.get_all_characters():
            char.tags.add(tag, category=category)

    def character_cleanup(self, char):
        """
        Clean all custom tags/attrs on a character.

        """
        if self.tagcategory:
            char.tags.remove(category=self.tagcategory)
            char.attributes.remove(category=self.tagcategory)

    def character_exit(self, char):
        """
        Have a character exit the room - return them to the room menu.

        """
        self.log(f"EXIT: {char} left room")
        from .menu import run_evscaperoom_menu

        self.character_cleanup(char)
        char.location = char.home

        # check if room should be deleted
        if len(self.get_all_characters()) < 1:
            self.delete()

        # we must run menu after deletion so we don't include this room!
        run_evscaperoom_menu(char)

    # Evennia hooks

    def at_object_receive(self, moved_obj, source_location, move_type="move", **kwargs):
        """
        Called when an object arrives in the room. This can be used to
        sum up the situation, set tags etc.

        """
        if utils.inherits_from(moved_obj, "evennia.objects.objects.DefaultCharacter"):
            self.log(f"JOIN: {moved_obj} joined room")
            self.state.character_enters(moved_obj)

    def at_object_leave(self, moved_obj, target_location, move_type="move", **kwargs):
        """
        Called when an object leaves the room; if this is a Character we need
        to clean them up and move them to the menu state.

        """
        if utils.inherits_from(moved_obj, "evennia.objects.objects.DefaultCharacter"):
            self.character_cleanup(moved_obj)
        if len(self.get_all_characters()) <= 1:
            # after this move there'll be no more characters in the room - delete the room!
            self.delete()
            # logger.log_info("DEBUG: Don't delete room when last player leaving")

    def delete(self):
        """
        Delete this room and all items related to it. Only move the players.

        """
        self.db.deleting = True
        for char in self.get_all_characters():
            self.character_exit(char)
        for obj in self.contents:
            obj.delete()
        self.log("END: Room cleaned up and deleted")
        return super().delete()

    def return_appearance(self, looker, **kwargs):
        obj, pos = self.get_position(looker)
        pos = (
            f"\n|x[{self.position_prep_map[pos]} on " f"{obj.get_display_name(looker)}]|n"
            if obj
            else ""
        )

        admin_only = ""
        if self.check_perm(looker, "Admin"):
            # only for admins
            objs = DefaultObject.objects.filter_family(db_location=self).exclude(id=looker.id)
            admin_only = "\n|xAdmin only: " + list_to_string(
                [obj.get_display_name(looker) for obj in objs]
            )

        return f"{self.db.desc}{pos}{admin_only}"



"""
Building menu tests.

"""

from evennia.commands.default.tests import BaseEvenniaCommandTest

from .building_menu import BuildingMenu, CmdNoMatch


class Submenu(BuildingMenu):
    def init(self, exit):
        self.add_choice("title", key="t", attr="key")


class TestBuildingMenu(BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        self.menu = BuildingMenu(caller=self.char1, obj=self.room1, title="test")
        self.menu.add_choice("title", key="t", attr="key")

    def test_quit(self):
        """Try to quit the building menu."""
        self.assertFalse(self.char1.cmdset.has("building_menu"))
        self.menu.open()
        self.assertTrue(self.char1.cmdset.has("building_menu"))
        self.call(CmdNoMatch(building_menu=self.menu), "q")
        # char1 tries to quit the editor
        self.assertFalse(self.char1.cmdset.has("building_menu"))

    def test_setattr(self):
        """Test the simple setattr provided by building menus."""
        self.menu.open()
        self.call(CmdNoMatch(building_menu=self.menu), "t")
        self.assertIsNotNone(self.menu.current_choice)
        self.call(CmdNoMatch(building_menu=self.menu), "some new title")
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.assertIsNone(self.menu.current_choice)
        self.assertEqual(self.room1.key, "some new title")
        self.call(CmdNoMatch(building_menu=self.menu), "q")

    def test_add_choice_without_key(self):
        """Try to add choices without keys."""
        choices = []
        for i in range(20):
            choices.append(self.menu.add_choice("choice", attr="test"))
        self.menu._add_keys_choice()
        keys = [
            "c",
            "h",
            "o",
            "i",
            "e",
            "ch",
            "ho",
            "oi",
            "ic",
            "ce",
            "cho",
            "hoi",
            "oic",
            "ice",
            "choi",
            "hoic",
            "oice",
            "choic",
            "hoice",
            "choice",
        ]
        for i in range(20):
            self.assertEqual(choices[i].key, keys[i])

        # Adding another key of the same title would break, no more available shortcut
        self.menu.add_choice("choice", attr="test")
        with self.assertRaises(ValueError):
            self.menu._add_keys_choice()

    def test_callbacks(self):
        """Test callbacks in menus."""
        self.room1.key = "room1"

        def on_enter(caller, menu):
            caller.msg("on_enter:{}".format(menu.title))

        def on_nomatch(caller, string, choice):
            caller.msg("on_nomatch:{},{}".format(string, choice.key))

        def on_leave(caller, obj):
            caller.msg("on_leave:{}".format(obj.key))

        self.menu.add_choice(
            "test", key="e", on_enter=on_enter, on_nomatch=on_nomatch, on_leave=on_leave
        )
        self.call(CmdNoMatch(building_menu=self.menu), "e", "on_enter:test")
        self.call(CmdNoMatch(building_menu=self.menu), "ok", "on_nomatch:ok,e")
        self.call(CmdNoMatch(building_menu=self.menu), "@", "on_leave:room1")
        self.call(CmdNoMatch(building_menu=self.menu), "q")

    def test_multi_level(self):
        """Test multi-level choices."""

        # Creaste three succeeding menu (t2 is contained in t1, t3 is contained in t2)
        def on_nomatch_t1(caller, menu):
            menu.move("whatever")  # this will be valid since after t1 is a joker

        def on_nomatch_t2(caller, menu):
            menu.move("t3")  # this time the key matters

        t1 = self.menu.add_choice("what", key="t1", on_nomatch=on_nomatch_t1)
        t2 = self.menu.add_choice("and", key="t1.*", on_nomatch=on_nomatch_t2)
        t3 = self.menu.add_choice("why", key="t1.*.t3")
        self.menu.open()

        # Move into t1
        self.assertIn(t1, self.menu.relevant_choices)
        self.assertNotIn(t2, self.menu.relevant_choices)
        self.assertNotIn(t3, self.menu.relevant_choices)
        self.assertIsNone(self.menu.current_choice)
        self.call(CmdNoMatch(building_menu=self.menu), "t1")
        self.assertEqual(self.menu.current_choice, t1)
        self.assertNotIn(t1, self.menu.relevant_choices)
        self.assertIn(t2, self.menu.relevant_choices)
        self.assertNotIn(t3, self.menu.relevant_choices)

        # Move into t2
        self.call(CmdNoMatch(building_menu=self.menu), "t2")
        self.assertEqual(self.menu.current_choice, t2)
        self.assertNotIn(t1, self.menu.relevant_choices)
        self.assertNotIn(t2, self.menu.relevant_choices)
        self.assertIn(t3, self.menu.relevant_choices)

        # Move into t3
        self.call(CmdNoMatch(building_menu=self.menu), "t3")
        self.assertEqual(self.menu.current_choice, t3)
        self.assertNotIn(t1, self.menu.relevant_choices)
        self.assertNotIn(t2, self.menu.relevant_choices)
        self.assertNotIn(t3, self.menu.relevant_choices)

        # Move back to t2
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.assertEqual(self.menu.current_choice, t2)
        self.assertNotIn(t1, self.menu.relevant_choices)
        self.assertNotIn(t2, self.menu.relevant_choices)
        self.assertIn(t3, self.menu.relevant_choices)

        # Move back into t1
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.assertEqual(self.menu.current_choice, t1)
        self.assertNotIn(t1, self.menu.relevant_choices)
        self.assertIn(t2, self.menu.relevant_choices)
        self.assertNotIn(t3, self.menu.relevant_choices)

        # Moves back to the main menu
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.assertIn(t1, self.menu.relevant_choices)
        self.assertNotIn(t2, self.menu.relevant_choices)
        self.assertNotIn(t3, self.menu.relevant_choices)
        self.assertIsNone(self.menu.current_choice)
        self.call(CmdNoMatch(building_menu=self.menu), "q")

    def test_submenu(self):
        """Test to add sub-menus."""

        def open_exit(menu):
            menu.open_submenu("evennia.contrib.base_systems.building_menu.tests.Submenu", self.exit)
            return False

        self.menu.add_choice("exit", key="x", on_enter=open_exit)
        self.menu.open()
        self.call(CmdNoMatch(building_menu=self.menu), "x")
        self.menu = self.char1.ndb._building_menu
        self.call(CmdNoMatch(building_menu=self.menu), "t")
        self.call(CmdNoMatch(building_menu=self.menu), "in")
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.call(CmdNoMatch(building_menu=self.menu), "@")
        self.menu = self.char1.ndb._building_menu
        self.assertEqual(self.char1.ndb._building_menu.obj, self.room1)
        self.call(CmdNoMatch(building_menu=self.menu), "q")
        self.assertEqual(self.exit.key, "in")



"""
Custom manager for HelpEntry objects.
"""
from django.db import IntegrityError

from evennia.server import signals
from evennia.typeclasses.managers import TypedObjectManager
from evennia.utils import logger, utils
from evennia.utils.utils import make_iter

__all__ = ("HelpEntryManager",)


class HelpEntryManager(TypedObjectManager):
    """
    This HelpEntryManager implements methods for searching
    and manipulating HelpEntries directly from the database.

    These methods will all return database objects
    (or QuerySets) directly.

    Evennia-specific:
    find_topicmatch
    find_apropos
    find_topicsuggestions
    find_topics_with_category
    all_to_category
    search_help (equivalent to evennia.search_helpentry)

    """

    def find_topicmatch(self, topicstr, exact=False):
        """
        Searches for matching topics or aliases based on player's
        input.

        Args:
            topcistr (str): Help topic to search for.
            exact (bool, optional): Require exact match
                (non-case-sensitive).  If `False` (default), match
                sub-parts of the string.

        Returns:
            matches (HelpEntries): Query results.

        """
        dbref = utils.dbref(topicstr)
        if dbref:
            return self.filter(id=dbref)
        topics = self.filter(db_key__iexact=topicstr)
        if not topics:
            topics = self.get_by_alias(topicstr)
        if not topics and not exact:
            topics = self.filter(db_key__istartswith=topicstr)
            if not topics:
                topics = self.filter(db_key__icontains=topicstr)
        return topics

    def find_apropos(self, topicstr):
        """
        Do a very loose search, returning all help entries containing
        the search criterion in their titles.

        Args:
            topicstr (str): Search criterion.

        Returns:
            matches (HelpEntries): Query results.

        """
        return self.filter(db_key__icontains=topicstr)

    def find_topicsuggestions(self, topicstr):
        """
        Do a fuzzy match, preferably within the category of the
        current topic.

        Args:
            topicstr (str): Search criterion.

        Returns:
            matches (Helpentries): Query results.

        """
        return self.filter(db_key__icontains=topicstr).exclude(db_key__iexact=topicstr)

    def find_topics_with_category(self, help_category):
        """
        Search topics having a particular category.

        Args:
            help_category (str): Category query criterion.

        Returns:
            matches (HelpEntries): Query results.

        """
        return self.filter(db_help_category__iexact=help_category)

    def get_all_topics(self):
        """
        Get all topics.

        Returns:
            all (HelpEntries): All topics.

        """
        return self.all()

    def get_all_categories(self):
        """
        Return all defined category names with at least one topic in
        them.

        Returns:
            matches (list): Unique list of category names across all
                topics.

        """
        return list(set(topic.help_category for topic in self.all()))

    def all_to_category(self, default_category):
        """
        Shifts all help entries in database to default_category.  This
        action cannot be reverted. It is used primarily by the engine
        when importing a default help database, making sure this ends
        up in one easily separated category.

        Args:
            default_category (str): Category to move entries to.

        """
        topics = self.all()
        for topic in topics:
            topic.help_category = default_category
            topic.save()
        string = "Help database moved to category {default_category}".format(
            default_category=default_category
        )
        logger.log_info(string)

    def search_help(self, ostring, help_category=None):
        """
        Retrieve a search entry object.

        Args:
            ostring (str): The help topic to look for.
            category (str): Limit the search to a particular help topic

        Returns:
            Queryset: An iterable with 0, 1 or more matches.

        """
        ostring = ostring.strip().lower()
        if help_category:
            return self.filter(db_key__iexact=ostring, db_help_category__iexact=help_category)
        else:
            return self.filter(db_key__iexact=ostring)

    def create_help(self, key, entrytext, category="General", locks=None, aliases=None, tags=None):
        """
        Create a static help entry in the help database. Note that Command
        help entries are dynamic and directly taken from the __doc__
        entries of the command. The database-stored help entries are
        intended for more general help on the game, more extensive info,
        in-game setting information and so on.

        Args:
            key (str): The name of the help entry.
            entrytext (str): The body of te help entry
            category (str, optional): The help category of the entry.
            locks (str, optional): A lockstring to restrict access.
            aliases (list of str, optional): List of alternative (likely shorter) keynames.
            tags (lst, optional): List of tags or tuples `(tag, category)`.

        Returns:
            help (HelpEntry): A newly created help entry.

        """
        try:
            new_help = self.model()
            new_help.key = key
            new_help.entrytext = entrytext
            new_help.help_category = category
            if locks:
                new_help.locks.add(locks)
            if aliases:
                new_help.aliases.add(make_iter(aliases))
            if tags:
                new_help.tags.batch_add(*tags)
            new_help.save()
            return new_help
        except IntegrityError:
            string = "Could not add help entry: key '%s' already exists." % key
            logger.log_err(string)
            return None
        except Exception:
            logger.log_trace()
            return None

        signals.SIGNAL_HELPENTRY_POST_CREATE.send(sender=new_help)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-06 17:31


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scripts", "0007_auto_20150403_2339")]

    operations = [
        migrations.AlterField(
            model_name="scriptdb",
            name="db_attributes",
            field=models.ManyToManyField(
                help_text="attributes on this object. An attribute can hold any pickle-able python object (see docs for special cases).",
                to="typeclasses.Attribute",
            ),
        ),
        migrations.AlterField(
            model_name="scriptdb",
            name="db_tags",
            field=models.ManyToManyField(
                help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                to="typeclasses.Tag",
            ),
        ),
    ]



"""
How to make a sword - example crafting tree for the crafting system.

See the `SwordSmithingBaseRecipe` in this module for an example of extendng the
recipe with a mocked 'skill' system (just random chance in our case). The skill
system used is game-specific but likely to be needed for most 'real' crafting
systems.

Note that 'tools' are references to the tools used - they don't need to be in
the inventory of the crafter. So when 'blast furnace' is given below, it is a
reference to a blast furnace used, not suggesting the crafter is carrying it
around with them.

## Sword crafting tree

::

    # base materials (consumables)

    iron ore, ash, sand, coal, oak wood, water, fur

    # base tools (marked with [T] for clarity and assumed to already exist)

    blast furnace[T], furnace[T], crucible[T], anvil[T],
    hammer[T], knife[T], cauldron[T]

    # recipes for making a sword

    pig iron = iron ore + 2xcoal + blast furnace[T]
    crucible_steel = pig iron + ash + sand + 2xcoal + crucible[T]
    sword blade = crucible steel + hammer[T] + anvil[T] + furnace[T]
    sword pommel = crucible steel + hammer[T] + anvil[T] + furnace[T]
    sword guard = crucible steel + hammer[T] + anvil[T] + furnace[T]

    rawhide = fur + knife[T]
    oak bark + cleaned oak wood = oak wood + knife[T]
    leather = rawhide + oak bark + water + cauldron[T]

    sword handle = cleaned oak wood + knife[T]

    sword = sword blade + sword guard + sword pommel
            + sword handle + leather + knife[T] + hammer[T] + furnace[T]


## Recipes used for spell casting

This is a simple example modifying the base Recipe to use as a way
to describe magical spells instead. It combines tools with
a skill (an attribute on the caster) in order to produce a magical effect.

The example `CmdCast` command can be added to the CharacterCmdset in
`mygame/commands/default_cmdsets` to test it out. The 'effects' are
just mocked for the example.

::
    # base tools (assumed to already exist)

    spellbook[T], wand[T]

    # skill (stored as Attribute on caster)

    firemagic skill level10+

    # recipe for fireball

    fireball = spellbook[T] + wand[T] + [firemagic skill lvl10+]

----

"""

from random import randint, random

from evennia.commands.command import Command, InterruptCommand

from .crafting import CraftingRecipe, CraftingValidationError, craft

# ------------------------------------------------------------
# Sword recipe
# ------------------------------------------------------------


class PigIronRecipe(CraftingRecipe):
    """
    Pig iron is a high-carbon result of melting iron in a blast furnace.

    """

    name = "pig iron"
    tool_tags = ["blast furnace"]
    consumable_tags = ["iron ore", "coal", "coal"]
    output_prototypes = [
        {
            "key": "Pig Iron ingot",
            "desc": "An ingot of crude pig iron.",
            "tags": [("pig iron", "crafting_material")],
        }
    ]


class CrucibleSteelRecipe(CraftingRecipe):
    """
    Mixing pig iron with impurities like ash and sand and melting it in a
    crucible produces a medieval level of steel (like damascus steel).

    """

    name = "crucible steel"
    tool_tags = ["crucible"]
    consumable_tags = ["pig iron", "ash", "sand", "coal", "coal"]
    output_prototypes = [
        {
            "key": "Crucible steel ingot",
            "desc": "An ingot of multi-colored crucible steel.",
            "tags": [("crucible steel", "crafting_material")],
        }
    ]


class _SwordSmithingBaseRecipe(CraftingRecipe):
    """
    A parent for all metallurgy sword-creation recipes. Those have a chance to
    failure but since steel is not lost in the process you can always try
    again.

    """

    success_message = "Your smithing work bears fruit and you craft {outputs}!"
    failed_message = (
        "You work and work but you are not happy with the result. You need to start over."
    )

    def craft(self, **kwargs):
        """
        Making a sword blade takes skill. Here we emulate this by introducing a
        random chance of failure (in a real game this could be a skill check
        against a skill found on `self.crafter`). In this case you can always
        start over since steel is not lost but can be re-smelted again for
        another try.

        Args:
            validated_inputs (list): all consumables/tools being used.
            **kwargs: any extra kwargs passed during crafting.

        Returns:
            any: The result of the craft, or None if a failure.

        Notes:
            Depending on if we return a crafting result from this
            method or not, `success_message` or `failure_message`
            will be echoed to the crafter.

            (for more control we could also message directly and raise
            crafting.CraftingError to abort craft process on failure).

        """
        if random.random() < 0.8:
            # 80% chance of success. This will spawn the sword and show
            # success-message.
            return super().craft(**kwargs)
        else:
            # fail and show failed message
            return None


class SwordBladeRecipe(_SwordSmithingBaseRecipe):
    """
    A [sword]blade requires hammering the steel out into shape using heat and
    force. This also includes the tang, which is the base for the hilt (the
    part of the sword you hold on to).

    """

    name = "sword blade"
    tool_tags = ["hammer", "anvil", "furnace"]
    consumable_tags = ["crucible steel"]
    output_prototypes = [
        {
            "key": "Sword blade",
            "desc": "A long blade that may one day become a sword.",
            "tags": [("sword blade", "crafting_material")],
        }
    ]


class SwordPommelRecipe(_SwordSmithingBaseRecipe):
    """
    The pommel is the 'button' or 'ball' etc the end of the sword hilt, holding
    it together.

    """

    name = "sword pommel"
    tool_tags = ["hammer", "anvil", "furnace"]
    consumable_tags = ["crucible steel"]
    output_prototypes = [
        {
            "key": "Sword pommel",
            "desc": "The pommel for a future sword.",
            "tags": [("sword pommel", "crafting_material")],
        }
    ]


class SwordGuardRecipe(_SwordSmithingBaseRecipe):
    """
    The guard stops the hand from accidentally sliding off the hilt onto the
    sword's blade and also protects the hand when parrying.

    """

    name = "sword guard"
    tool_tags = ["hammer", "anvil", "furnace"]
    consumable_tags = ["crucible steel"]
    output_prototypes = [
        {
            "key": "Sword guard",
            "desc": "The cross-guard for a future sword.",
            "tags": [("sword guard", "crafting_material")],
        }
    ]


class RawhideRecipe(CraftingRecipe):
    """
    Rawhide is animal skin cleaned and stripped of hair.

    """

    name = "rawhide"
    tool_tags = ["knife"]
    consumable_tags = ["fur"]
    output_prototypes = [
        {
            "key": "Rawhide",
            "desc": "Animal skin, cleaned and with hair removed.",
            "tags": [("rawhide", "crafting_material")],
        }
    ]


class OakBarkRecipe(CraftingRecipe):
    """
    The actual thing needed for tanning leather is Tannin, but we skip
    the step of refining tannin from the bark and use the bark as-is.

    This produces two outputs - the bark and the cleaned wood.
    """

    name = "oak bark"
    tool_tags = ["knife"]
    consumable_tags = ["oak wood"]
    output_prototypes = [
        {
            "key": "Oak bark",
            "desc": "Bark of oak, stripped from the core wood.",
            "tags": [("oak bark", "crafting_material")],
        },
        {
            "key": "Oak Wood (cleaned)",
            "desc": "Oakwood core, stripped of bark.",
            "tags": [("cleaned oak wood", "crafting_material")],
        },
    ]


class LeatherRecipe(CraftingRecipe):
    """
    Leather is produced by tanning rawhide in a process traditionally involving
    the chemical Tannin. Here we abbreviate this process a bit. Maybe a
    'tanning rack' tool should be required too ...

    """

    name = "leather"
    tool_tags = ["cauldron"]
    consumable_tags = ["rawhide", "oak bark", "water"]
    output_prototypes = [
        {
            "key": "Piece of Leather",
            "desc": "A piece of leather.",
            "tags": [("leather", "crafting_material")],
        }
    ]


class SwordHandleRecipe(CraftingRecipe):
    """
    The handle is the part of the hilt between the guard and the pommel where
    you hold the sword. It consists of wooden pieces around the steel tang. It
    is wrapped in leather, but that will be added at the end.

    """

    name = "sword handle"
    tool_tags = ["knife"]
    consumable_tags = ["cleaned oak wood"]
    output_prototypes = [
        {
            "key": "Sword handle",
            "desc": "Two pieces of wood to be be fitted onto a sword's tang as its handle.",
            "tags": [("sword handle", "crafting_material")],
        }
    ]


class SwordRecipe(_SwordSmithingBaseRecipe):
    """
    A finished sword consists of a Blade ending in a non-sharp part called the
    Tang. The cross Guard is put over the tang against the edge of the blade.
    The Handle is put over the tang to give something easier to hold. The
    Pommel locks everything in place. The handle is wrapped in leather
    strips for better grip.

    This covers only a single 'sword' type.

    """

    name = "sword"
    tool_tags = ["hammer", "furnace", "knife"]
    consumable_tags = ["sword blade", "sword guard", "sword pommel", "sword handle", "leather"]
    output_prototypes = [
        {
            "key": "Sword",
            "desc": "A bladed weapon.",
            # setting the tag as well - who knows if one can make something from this too!
            "tags": [("sword", "crafting_material")],
        }
        # obviously there would be other properties of a 'sword' added here
        # too, depending on how combat works in the your game!
    ]
    # this requires more precision
    exact_consumable_order = True


# ------------------------------------------------------------
# Recipes for spell casting
# ------------------------------------------------------------


class _MagicRecipe(CraftingRecipe):
    """
    A base 'recipe' to represent magical spells.

    We *could* treat this just like the sword above - by combining the wand and spellbook to make a
    fireball object that the user can then throw with another command. For this example we instead
    generate 'magical effects' as strings+values that we would then supposedly inject into a
    combat system or other resolution system.

    We also assume that the crafter has skills set on itself as plain Attributes.

    """

    name = ""
    # all spells require a spellbook and a wand (so there!)
    tool_tags = ["spellbook", "wand"]

    error_tool_missing_message = "Cannot cast spells without {missing}."
    success_message = "You successfully cast the spell!"
    # custom properties
    skill_requirement = []  # this should be on the form [(skillname, min_level)]
    skill_roll = ""  # skill to roll for success
    desired_effects = []  # on the form [(effect, value), ...]
    failure_effects = []  # ''
    error_too_low_skill_level = "Your skill {skill_name} is too low to cast {spell}."
    error_no_skill_roll = "You must have the skill {skill_name} to cast the spell {spell}."

    def pre_craft(self, **kwargs):
        """
        This is where we do input validation. We want to do the
        normal validation of the tools, but also check for a skill
        on the crafter. This must set the result on `self.validated_inputs`.
        We also set the crafter's relevant skill value on `self.skill_roll_value`.

        Args:
            **kwargs: Any optional extra kwargs passed during initialization of
                the recipe class.

        Raises:
            CraftingValidationError: If validation fails. At this point the crafter
                is expected to have been informed of the problem already.

        """
        # this will check so the spellbook and wand are at hand.
        super().pre_craft(**kwargs)

        # at this point we have the items available, let's also check for the skill. We
        # assume the  crafter has the skill available as an Attribute
        # on itself.

        crafter = self.crafter
        for skill_name, min_value in self.skill_requirements:
            skill_value = crafter.attributes.get(skill_name)

            if skill_value is None or skill_value < min_value:
                self.msg(
                    self.error_too_low_skill_level.format(skill_name=skill_name, spell=self.name)
                )
                raise CraftingValidationError

        # get the value of the skill to roll
        self.skill_roll_value = self.crafter.attributes.get(self.skill_roll)
        if self.skill_roll_value is None:
            self.msg(self.error_no_skill_roll.format(skill_name=self.skill_roll, spell=self.name))
            raise CraftingValidationError

    def do_craft(self, **kwargs):
        """
        'Craft' the magical effect. When we get to this point we already know we have all the
        prequisite for creating the effect. In this example we will store the effect on the crafter;
        maybe this enhances the crafter or makes a new attack available to them in combat.

        An alternative to this would of course be to spawn an actual object for the effect, like
        creating a potion or an actual fireball-object to throw (this depends on how your combat
        works).

        """
        # we do a simple skill check here.
        if randint(1, 18) <= self.skill_roll_value:
            # a success!
            return True, self.desired_effects
        else:
            # a failure!
            return False, self.failure_effects

    def post_craft(self, craft_result, **kwargs):
        """
        Always called at the end of crafting, regardless of successful or not.

        Since we get a custom craft result (True/False, effects) we need to
        wrap the original post_craft to output the error messages for us
        correctly.

        """
        success = False
        if craft_result:
            success, _ = craft_result
        # default post_craft just checks if craft_result is truthy or not.
        # we don't care about its return value since we already have craft_result.
        super().post_craft(success, **kwargs)
        return craft_result


class FireballRecipe(_MagicRecipe):
    """
    A Fireball is a magical effect that can be thrown at a target to cause damage.

    Note that the magic-effects are just examples, an actual rule system would
    need to be created to understand what they mean when used.

    """

    name = "fireball"
    skill_requirements = [("firemagic", 10)]  # skill 'firemagic' lvl 10 or higher
    skill_roll = "firemagic"
    success_message = "A ball of flame appears!"
    desired_effects = [("target_fire_damage", 25), ("ranged_attack", -2), ("mana_cost", 12)]
    failure_effects = [("self_fire_damage", 5), ("mana_cost", 5)]


class HealingRecipe(_MagicRecipe):
    """
    Healing magic will restore a certain amount of health to the target over time.

    Note that the magic-effects are just examples, an actual rule system would
    need to be created to understand what they mean.

    """

    name = "heal"
    skill_requirements = [("bodymagic", 5), ("empathy", 10)]
    skill_roll = "bodymagic"
    success_message = "You successfully extend your healing aura."
    desired_effects = [("healing", 15), ("mana_cost", 5)]
    failure_effects = []


class CmdCast(Command):
    """
    Cast a magical spell.

    Usage:
        cast <spell> <target>

    """

    key = "cast"

    def parse(self):
        """
        Simple parser, assuming spellname doesn't have spaces.
        Stores result in self.target and self.spellname.

        """
        args = self.args.strip().lower()
        target = None
        if " " in args:
            self.spellname, *target = args.split(" ", 1)
        else:
            self.spellname = args

        if not self.spellname:
            self.caller.msg("You must specify a spell name.")
            raise InterruptCommand

        if target:
            self.target = self.caller.search(target[0].strip())
            if not self.target:
                raise InterruptCommand
        else:
            self.target = self.caller

    def func(self):
        # all items carried by the caller could work
        possible_tools = self.caller.contents

        try:
            # if this completes without an exception, the caster will have
            # a new magic_effect set on themselves, ready to use or apply in some way.
            success, effects = craft(
                self.caller, self.spellname, *possible_tools, raise_exception=True
            )
        except CraftingValidationError:
            return
        except KeyError:
            self.caller.msg(f"You don't know of a spell called '{self.spellname}'")
            return

        # Applying the magical effect to target would happen below.
        # self.caller.db.active_spells[self.spellname] holds all the effects
        # of this particular prepared spell. For a fireball you could perform
        # an attack roll here and apply damage if you hit. For healing you would heal the target
        # (which could be yourself) by a number of health points given by the recipe.
        effect_txt = ", ".join(f"{eff[0]}({eff[1]})" for eff in effects)
        success_txt = "|gsucceeded|n" if success else "|rfailed|n"
        self.caller.msg(
            f"Casting the spell {self.spellname} on {self.target} {success_txt}, "
            f"causing the following effects: {effect_txt}."
        )



"""

System commands

"""


import code
import datetime
import os
import sys
import time
import traceback

import django
import twisted
from django.conf import settings

import evennia
from evennia.accounts.models import AccountDB
from evennia.scripts.taskhandler import TaskHandlerTask
from evennia.utils import gametime, logger, search, utils
from evennia.utils.eveditor import EvEditor
from evennia.utils.evmenu import ask_yes_no
from evennia.utils.evtable import EvTable
from evennia.utils.utils import class_from_module, iter_to_str

COMMAND_DEFAULT_CLASS = class_from_module(settings.COMMAND_DEFAULT_CLASS)
_TASK_HANDLER = None
_BROADCAST_SERVER_RESTART_MESSAGES = settings.BROADCAST_SERVER_RESTART_MESSAGES

# delayed imports
_RESOURCE = None
_IDMAPPER = None

# limit symbol import for API
__all__ = (
    "CmdAccounts",
    "CmdReload",
    "CmdReset",
    "CmdShutdown",
    "CmdPy",
    "CmdService",
    "CmdAbout",
    "CmdTime",
    "CmdServerLoad",
    "CmdTasks",
    "CmdTickers",
)


class CmdReload(COMMAND_DEFAULT_CLASS):
    """
    reload the server

    Usage:
      reload [reason]

    This restarts the server. The Portal is not
    affected. Non-persistent scripts will survive a reload (use
    reset to purge) and at_reload() hooks will be called.
    """

    key = "@reload"
    aliases = ["@restart"]
    locks = "cmd:perm(reload) or perm(Developer)"
    help_category = "System"

    def func(self):
        """
        Reload the system.
        """
        reason = ""
        if self.args:
            reason = "(Reason: %s) " % self.args.rstrip(".")
        if _BROADCAST_SERVER_RESTART_MESSAGES:
            evennia.SESSION_HANDLER.announce_all(f" Server restart initiated {reason}...")
        evennia.SESSION_HANDLER.portal_restart_server()


class CmdReset(COMMAND_DEFAULT_CLASS):
    """
    reset and reboot the server

    Usage:
      reset

    Notes:
      For normal updating you are recommended to use reload rather
      than this command. Use shutdown for a complete stop of
      everything.

    This emulates a cold reboot of the Server component of Evennia.
    The difference to shutdown is that the Server will auto-reboot
    and that it does not affect the Portal, so no users will be
    disconnected. Contrary to reload however, all shutdown hooks will
    be called and any non-database saved scripts, ndb-attributes,
    cmdsets etc will be wiped.

    """

    key = "@reset"
    aliases = ["@reboot"]
    locks = "cmd:perm(reload) or perm(Developer)"
    help_category = "System"

    def func(self):
        """
        Reload the system.
        """
        evennia.SESSION_HANDLER.announce_all(" Server resetting/restarting ...")
        evennia.SESSION_HANDLER.portal_reset_server()


class CmdShutdown(COMMAND_DEFAULT_CLASS):

    """
    stop the server completely

    Usage:
      shutdown [announcement]

    Gracefully shut down both Server and Portal.
    """

    key = "@shutdown"
    locks = "cmd:perm(shutdown) or perm(Developer)"
    help_category = "System"

    def func(self):
        """Define function"""
        # Only allow shutdown if caller has session
        if not self.caller.sessions.get():
            return
        self.msg("Shutting down server ...")
        announcement = "\nServer is being SHUT DOWN!\n"
        if self.args:
            announcement += "%s\n" % self.args
        logger.log_info(f"Server shutdown by {self.caller.name}.")
        evennia.SESSION_HANDLER.announce_all(announcement)
        evennia.SESSION_HANDLER.portal_shutdown()


def _py_load(caller):
    return ""


def _py_code(caller, buf):
    """
    Execute the buffer.
    """
    measure_time = caller.db._py_measure_time
    client_raw = caller.db._py_clientraw
    string = "Executing code%s ..." % (" (measure timing)" if measure_time else "")
    caller.msg(string)
    _run_code_snippet(
        caller, buf, mode="exec", measure_time=measure_time, client_raw=client_raw, show_input=False
    )
    return True


def _py_quit(caller):
    del caller.db._py_measure_time
    caller.msg("Exited the code editor.")


def _run_code_snippet(
    caller, pycode, mode="eval", measure_time=False, client_raw=False, show_input=True
):
    """
    Run code and try to display information to the caller.

    Args:
        caller (Object): The caller.
        pycode (str): The Python code to run.
        measure_time (bool, optional): Should we measure the time of execution?
        client_raw (bool, optional): Should we turn off all client-specific escaping?
        show_input (bookl, optional): Should we display the input?

    """
    # Try to retrieve the session
    if hasattr(caller, "sessions"):
        sessions = caller.sessions.all()

    available_vars = evennia_local_vars(caller)

    if show_input:
        for session in sessions:
            data = {
                # TODO: 'highlight' is not used yet
                "text": (f">>> {pycode}", {"type": "py_input"}),
                "options": {"raw": True, "highlight": True},
            }
            try:
                caller.msg(session=session, **data)
            except TypeError:
                caller.msg(**data)

    try:
        # reroute standard output to game client console
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        class FakeStd:
            def __init__(self, caller):
                self.caller = caller

            def write(self, string):
                self.caller.msg(text=(string.rstrip("\n"), {"type": "py_output"}))

        fake_std = FakeStd(caller)
        sys.stdout = fake_std
        sys.stderr = fake_std

        try:
            pycode_compiled = compile(pycode, "", mode)
        except Exception:
            mode = "exec"
            pycode_compiled = compile(pycode, "", mode)

        duration = ""
        if measure_time:
            t0 = time.time()
            ret = eval(pycode_compiled, {}, available_vars)
            t1 = time.time()
            duration = f" (runtime ~ {(t1 - t0) * 1000:.4f} ms)"
            caller.msg(duration)
        else:
            ret = eval(pycode_compiled, {}, available_vars)

    except Exception:
        errlist = traceback.format_exc().split("\n")
        if len(errlist) > 4:
            errlist = errlist[4:]
        ret = "\n".join("%s" % line for line in errlist if line)
    finally:
        # return to old stdout
        sys.stdout = old_stdout
        sys.stderr = old_stderr

    if ret is None:
        return

    if not client_raw:
        ret = str(ret)

    if isinstance(ret, tuple):
        # we must convert here to allow msg to pass it (a tuple is confused
        # with a outputfunc structure)
        ret = str(ret)

    for session in sessions:
        try:
            caller.msg(
                (ret, {"type": "py_output"}),
                session=session,
                options={"raw": True, "client_raw": client_raw, "highlight": True},
            )
        except TypeError:
            caller.msg(
                (ret, {"type": "py_output"}),
                options={"raw": True, "client_raw": client_raw, "highlight": True},
            )


def evennia_local_vars(caller):
    """Return Evennia local variables usable in the py command as a dictionary."""
    import evennia

    return {
        "self": caller,
        "me": caller,
        "here": getattr(caller, "location", None),
        "evennia": evennia,
        "ev": evennia,
        "inherits_from": utils.inherits_from,
    }


class EvenniaPythonConsole(code.InteractiveConsole):

    """Evennia wrapper around a Python interactive console."""

    def __init__(self, caller):
        super().__init__(evennia_local_vars(caller))
        self.caller = caller

    def write(self, string):
        """Don't send to stderr, send to self.caller."""
        self.caller.msg(string)

    def push(self, line):
        """Push some code, whether complete or not."""
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        class FakeStd:
            def __init__(self, caller):
                self.caller = caller

            def write(self, string):
                self.caller.msg(string.split("\n", 1)[0])

        fake_std = FakeStd(self.caller)
        sys.stdout = fake_std
        sys.stderr = fake_std
        result = None
        try:
            result = super().push(line)
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr
        return result


class CmdPy(COMMAND_DEFAULT_CLASS):
    """
    execute a snippet of python code

    Usage:
      py [cmd]
      py/edit
      py/time <cmd>
      py/clientraw <cmd>
      py/noecho

    Switches:
      time - output an approximate execution time for <cmd>
      edit - open a code editor for multi-line code experimentation
      clientraw - turn off all client-specific escaping. Note that this may
        lead to different output depending on prototocol (such as angular brackets
        being parsed as HTML in the webclient but not in telnet clients)
      noecho - in Python console mode, turn off the input echo (e.g. if your client
        does this for you already)

    Without argument, open a Python console in-game. This is a full console,
    accepting multi-line Python code for testing and debugging. Type `exit()` to
    return to the game. If Evennia is reloaded, the console will be closed.

    Enter a line of instruction after the 'py' command to execute it
    immediately.  Separate multiple commands by ';' or open the code editor
    using the /edit switch (all lines added in editor will be executed
    immediately when closing or using the execute command in the editor).

    A few variables are made available for convenience in order to offer access
    to the system (you can import more at execution time).

    Available variables in py environment:
      self, me                   : caller
      here                       : caller.location
      evennia                    : the evennia API
      inherits_from(obj, parent) : check object inheritance

    You can explore The evennia API from inside the game by calling
    the `__doc__` property on entities:
        py evennia.__doc__
        py evennia.managers.__doc__

    |rNote: In the wrong hands this command is a severe security risk.  It
    should only be accessible by trusted server admins/superusers.|n

    """

    key = "@py"
    aliases = ["@!"]
    switch_options = ("time", "edit", "clientraw", "noecho")
    locks = "cmd:perm(py) or perm(Developer)"
    help_category = "System"
    arg_regex = ""

    def func(self):
        """hook function"""

        caller = self.caller
        pycode = self.args

        noecho = "noecho" in self.switches

        if "edit" in self.switches:
            caller.db._py_measure_time = "time" in self.switches
            caller.db._py_clientraw = "clientraw" in self.switches
            EvEditor(
                self.caller,
                loadfunc=_py_load,
                savefunc=_py_code,
                quitfunc=_py_quit,
                key="Python exec: :w  or :!",
                persistent=True,
                codefunc=_py_code,
            )
            return

        if not pycode:
            # Run in interactive mode
            console = EvenniaPythonConsole(self.caller)
            banner = (
                "|gEvennia Interactive Python mode{echomode}\n"
                "Python {version} on {platform}".format(
                    echomode=" (no echoing of prompts)" if noecho else "",
                    version=sys.version,
                    platform=sys.platform,
                )
            )
            self.msg(banner)
            line = ""
            main_prompt = "|x[py mode - quit() to exit]|n"
            prompt = main_prompt
            while line.lower() not in ("exit", "exit()"):
                try:
                    line = yield (prompt)
                    if noecho:
                        prompt = "..." if console.push(line) else main_prompt
                    else:
                        if line:
                            self.caller.msg(f">>> {line}")
                        prompt = line if console.push(line) else main_prompt
                except SystemExit:
                    break
            self.msg("|gClosing the Python console.|n")
            return

        _run_code_snippet(
            caller,
            self.args,
            measure_time="time" in self.switches,
            client_raw="clientraw" in self.switches,
        )


class CmdAccounts(COMMAND_DEFAULT_CLASS):
    """
    Manage registered accounts

    Usage:
      accounts [nr]
      accounts/delete <name or #id> [: reason]

    Switches:
      delete    - delete an account from the server

    By default, lists statistics about the Accounts registered with the game.
    It will list the <nr> amount of latest registered accounts
    If not given, <nr> defaults to 10.
    """

    key = "@accounts"
    aliases = ["@account"]
    switch_options = ("delete",)
    locks = "cmd:perm(listaccounts) or perm(Admin)"
    help_category = "System"

    def func(self):
        """List the accounts"""

        caller = self.caller
        args = self.args

        if "delete" in self.switches:
            account = getattr(caller, "account")
            if not account or not account.check_permstring("Developer"):
                caller.msg("You are not allowed to delete accounts.")
                return
            if not args:
                caller.msg("Usage: accounts/delete <name or #id> [: reason]")
                return
            reason = ""
            if ":" in args:
                args, reason = [arg.strip() for arg in args.split(":", 1)]
            # We use account_search since we want to be sure to find also accounts
            # that lack characters.
            accounts = search.account_search(args)
            if not accounts:
                self.msg("Could not find an account by that name.")
                return
            if len(accounts) > 1:
                string = "There were multiple matches:\n"
                string += "\n".join(" %s %s" % (account.id, account.key) for account in accounts)
                self.msg(string)
                return
            account = accounts.first()
            if not account.access(caller, "delete"):
                self.msg("You don't have the permissions to delete that account.")
                return
            username = account.username
            # ask for confirmation
            confirm = (
                "It is often better to block access to an account rather than to delete it. "
                "|yAre you sure you want to permanently delete "
                "account '|n{}|y'|n yes/[no]?".format(username)
            )
            answer = yield (confirm)
            if answer.lower() not in ("y", "yes"):
                caller.msg("Canceled deletion.")
                return

            # Boot the account then delete it.
            self.msg("Informing and disconnecting account ...")
            string = f"\nYour account '{username}' is being *permanently* deleted.\n"
            if reason:
                string += " Reason given:\n  '%s'" % reason
            account.msg(string)
            logger.log_sec(
                f"Account Deleted: {account} (Reason: {reason}, Caller: {caller}, IP:"
                f" {self.session.address})."
            )
            account.delete()
            self.msg("Account %s was successfully deleted." % username)
            return

        # No switches, default to displaying a list of accounts.
        if self.args and self.args.isdigit():
            nlim = int(self.args)
        else:
            nlim = 10

        naccounts = AccountDB.objects.count()

        # typeclass table
        dbtotals = AccountDB.objects.object_totals()
        typetable = self.styled_table(
            "|wtypeclass|n", "|wcount|n", "|w%%|n", border="cells", align="l"
        )
        for path, count in dbtotals.items():
            typetable.add_row(path, count, "%.2f" % ((float(count) / naccounts) * 100))
        # last N table
        plyrs = AccountDB.objects.all().order_by("db_date_created")[max(0, naccounts - nlim) :]
        latesttable = self.styled_table(
            "|wcreated|n", "|wdbref|n", "|wname|n", "|wtypeclass|n", border="cells", align="l"
        )
        for ply in plyrs:
            latesttable.add_row(
                utils.datetime_format(ply.date_created), ply.dbref, ply.key, ply.path
            )

        string = f"\n|wAccount typeclass distribution:|n\n{typetable}"
        string += f"\n|wLast {min(naccounts, nlim)} Accounts created:|n\n{latesttable}"
        caller.msg(string)


class CmdService(COMMAND_DEFAULT_CLASS):
    """
    manage system services

    Usage:
      service[/switch] <service>

    Switches:
      list   - shows all available services (default)
      start  - activates or reactivate a service
      stop   - stops/inactivate a service (can often be restarted)
      delete - tries to permanently remove a service

    Service management system. Allows for the listing,
    starting, and stopping of services. If no switches
    are given, services will be listed. Note that to operate on the
    service you have to supply the full (green or red) name as given
    in the list.
    """

    key = "@service"
    aliases = ["@services"]
    switch_options = ("list", "start", "stop", "delete")
    locks = "cmd:perm(service) or perm(Developer)"
    help_category = "System"

    def func(self):
        """Implement command"""

        caller = self.caller
        switches = self.switches

        if switches and switches[0] not in ("list", "start", "stop", "delete"):
            caller.msg("Usage: service/<list|start|stop|delete> [servicename]")
            return

        # get all services
        service_collection = evennia.EVENNIA_SERVER_SERVICE

        if not switches or switches[0] == "list":
            # Just display the list of installed services and their
            # status, then exit.
            table = self.styled_table(
                "|wService|n (use services/start|stop|delete)", "|wstatus", align="l"
            )
            for service in service_collection.services:
                table.add_row(service.name, service.running and "|gRunning" or "|rNot Running")
            caller.msg(str(table))
            return

        # Get the service to start / stop

        try:
            service = service_collection.getServiceNamed(self.args)
        except Exception:
            string = "Invalid service name. This command is case-sensitive. "
            string += "See service/list for valid service name (enter the full name exactly)."
            caller.msg(string)
            return

        if switches[0] in ("stop", "delete"):
            # Stopping/killing a service gracefully closes it and disconnects
            # any connections (if applicable).

            delmode = switches[0] == "delete"
            if not service.running:
                caller.msg("That service is not currently running.")
                return
            if service.name[:7] == "Evennia":
                if delmode:
                    caller.msg("You cannot remove a core Evennia service (named 'Evennia*').")
                    return
                string = (
                    "|RYou seem to be shutting down a core Evennia "
                    "service (named 'Evennia*').\nNote that stopping "
                    "some TCP port services will *not* disconnect users "
                    "*already* connected on those ports, but *may* "
                    "instead cause spurious errors for them.\nTo safely "
                    "and permanently remove ports, change settings file "
                    "and restart the server.|n\n"
                )
                caller.msg(string)

            if delmode:
                service.stopService()
                service_collection.removeService(service)
                caller.msg(f"|gStopped and removed service '{self.args}'.|n")
            else:
                caller.msg(f"Stopping service '{self.args}'...")
                try:
                    service.stopService()
                except Exception as err:
                    caller.msg(
                        f"|rErrors were reported when stopping this service{err}.\n"
                        "If there are remaining problems, try reloading "
                        "or rebooting the server."
                    )
                caller.msg(f"|g... Stopped service '{self.args}'.|n")
            return

        if switches[0] == "start":
            # Attempt to start a service.
            if service.running:
                caller.msg("That service is already running.")
                return
            caller.msg(f"Starting service '{self.args}' ...")
            try:
                service.startService()
            except Exception as err:
                caller.msg(
                    f"|rErrors were reported when starting this service{err}.\n"
                    "If there are remaining problems, try reloading the server, changing the "
                    "settings if it's a non-standard service.|n"
                )
            caller.msg("|gService started.|n")


class CmdAbout(COMMAND_DEFAULT_CLASS):
    """
    show Evennia info

    Usage:
      about

    Display info about the game engine.
    """

    key = "@about"
    aliases = "@version"
    locks = "cmd:all()"
    help_category = "System"

    def func(self):
        """Display information about server or target"""

        string = """
         |cEvennia|n MU* development system

         |wEvennia version|n: {version}
         |wOS|n: {os}
         |wPython|n: {python}
         |wTwisted|n: {twisted}
         |wDjango|n: {django}

         |wHomepage|n https://evennia.com
         |wCode|n https://github.com/evennia/evennia
         |wDemo|n https://demo.evennia.com
         |wGame listing|n https://games.evennia.com
         |wChat|n https://discord.gg/AJJpcRUhtF
         |wForum|n https://github.com/evennia/evennia/discussions
         |wLicence|n https://opensource.org/licenses/BSD-3-Clause
         |wMaintainer|n (2010-)   Griatch (griatch AT gmail DOT com)
         |wMaintainer|n (2006-10) Greg Taylor

        """.format(
            version=utils.get_evennia_version(),
            os=os.name,
            python=sys.version.split()[0],
            twisted=twisted.version.short(),
            django=django.get_version(),
        )
        self.msg(string)


class CmdTime(COMMAND_DEFAULT_CLASS):
    """
    show server time statistics

    Usage:
      time

    List Server time statistics such as uptime
    and the current time stamp.
    """

    key = "@time"
    aliases = "@uptime"
    locks = "cmd:perm(time) or perm(Player)"
    help_category = "System"

    def func(self):
        """Show server time data in a table."""
        table1 = self.styled_table("|wServer time", "", align="l", width=78)
        table1.add_row("Current uptime", utils.time_format(gametime.uptime(), 3))
        table1.add_row("Portal uptime", utils.time_format(gametime.portal_uptime(), 3))
        table1.add_row("Total runtime", utils.time_format(gametime.runtime(), 2))
        table1.add_row("First start", datetime.datetime.fromtimestamp(gametime.server_epoch()))
        table1.add_row("Current time", datetime.datetime.now())
        table1.reformat_column(0, width=30)
        table2 = self.styled_table(
            "|wIn-Game time",
            "|wReal time x %g" % gametime.TIMEFACTOR,
            align="l",
            width=78,
            border_top=0,
        )
        epochtxt = "Epoch (%s)" % ("from settings" if settings.TIME_GAME_EPOCH else "server start")
        table2.add_row(epochtxt, datetime.datetime.fromtimestamp(gametime.game_epoch()))
        table2.add_row("Total time passed:", utils.time_format(gametime.gametime(), 2))
        table2.add_row(
            "Current time ", datetime.datetime.fromtimestamp(gametime.gametime(absolute=True))
        )
        table2.reformat_column(0, width=30)
        self.msg(str(table1) + "\n" + str(table2))


class CmdServerLoad(COMMAND_DEFAULT_CLASS):
    """
    show server load and memory statistics

    Usage:
       server[/mem]

    Switches:
        mem - return only a string of the current memory usage
        flushmem - flush the idmapper cache

    This command shows server load statistics and dynamic memory
    usage. It also allows to flush the cache of accessed database
    objects.

    Some Important statistics in the table:

    |wServer load|n is an average of processor usage. It's usually
    between 0 (no usage) and 1 (100% usage), but may also be
    temporarily higher if your computer has multiple CPU cores.

    The |wResident/Virtual memory|n displays the total memory used by
    the server process.

    Evennia |wcaches|n all retrieved database entities when they are
    loaded by use of the idmapper functionality. This allows Evennia
    to maintain the same instances of an entity and allowing
    non-persistent storage schemes. The total amount of cached objects
    are displayed plus a breakdown of database object types.

    The |wflushmem|n switch allows to flush the object cache. Please
    note that due to how Python's memory management works, releasing
    caches may not show you a lower Residual/Virtual memory footprint,
    the released memory will instead be re-used by the program.

    """

    key = "@server"
    aliases = ["@serverload"]
    switch_options = ("mem", "flushmem")
    locks = "cmd:perm(list) or perm(Developer)"
    help_category = "System"

    def func(self):
        """Show list."""

        global _IDMAPPER
        if not _IDMAPPER:
            from evennia.utils.idmapper import models as _IDMAPPER

        if "flushmem" in self.switches:
            # flush the cache
            prev, _ = _IDMAPPER.cache_size()
            nflushed = _IDMAPPER.flush_cache()
            now, _ = _IDMAPPER.cache_size()
            string = (
                "The Idmapper cache freed |w{idmapper}|n database objects.\n"
                "The Python garbage collector freed |w{gc}|n Python instances total."
            )
            self.msg(string.format(idmapper=(prev - now), gc=nflushed))
            return

        # display active processes

        os_windows = os.name == "nt"
        pid = os.getpid()

        if os_windows:
            # Windows requires the psutil module to even get paltry
            # statistics like this (it's pretty much worthless,
            # unfortunately, since it's not specific to the process) /rant
            try:
                import psutil

                has_psutil = True
            except ImportError:
                has_psutil = False

            if has_psutil:
                loadavg = psutil.cpu_percent()
                _mem = psutil.virtual_memory()
                rmem = _mem.used / (1000.0 * 1000)
                pmem = _mem.percent

                if "mem" in self.switches:
                    string = "Total computer memory usage: |w%g|n MB (%g%%)"
                    self.msg(string % (rmem, pmem))
                    return
                # Display table
                loadtable = self.styled_table("property", "statistic", align="l")
                loadtable.add_row("Total CPU load", "%g %%" % loadavg)
                loadtable.add_row("Total computer memory usage", "%g MB (%g%%)" % (rmem, pmem))
                loadtable.add_row("Process ID", "%g" % pid),
            else:
                loadtable = (
                    "Not available on Windows without 'psutil' library "
                    "(install with |wpip install psutil|n)."
                )

        else:
            # Linux / BSD (OSX) - proper pid-based statistics

            global _RESOURCE
            if not _RESOURCE:
                import resource as _RESOURCE

            loadavg = os.getloadavg()[0]
            rmem = (
                float(os.popen("ps -p %d -o %s | tail -1" % (pid, "rss")).read()) / 1000.0
            )  # resident memory
            vmem = (
                float(os.popen("ps -p %d -o %s | tail -1" % (pid, "vsz")).read()) / 1000.0
            )  # virtual memory
            pmem = float(
                os.popen("ps -p %d -o %s | tail -1" % (pid, "%mem")).read()
            )  # % of resident memory to total
            rusage = _RESOURCE.getrusage(_RESOURCE.RUSAGE_SELF)

            if "mem" in self.switches:
                string = "Memory usage: RMEM: |w%g|n MB (%g%%), VMEM (res+swap+cache): |w%g|n MB."
                self.msg(string % (rmem, pmem, vmem))
                return

            loadtable = self.styled_table("property", "statistic", align="l")
            loadtable.add_row("Server load (1 min)", "%g" % loadavg)
            loadtable.add_row("Process ID", "%g" % pid),
            loadtable.add_row("Memory usage", "%g MB (%g%%)" % (rmem, pmem))
            loadtable.add_row("Virtual address space", "")
            loadtable.add_row("|x(resident+swap+caching)|n", "%g MB" % vmem)
            loadtable.add_row(
                "CPU time used (total)",
                "%s (%gs)" % (utils.time_format(rusage.ru_utime), rusage.ru_utime),
            )
            loadtable.add_row(
                "CPU time used (user)",
                "%s (%gs)" % (utils.time_format(rusage.ru_stime), rusage.ru_stime),
            )
            loadtable.add_row(
                "Page faults",
                "%g hard,  %g soft, %g swapouts"
                % (rusage.ru_majflt, rusage.ru_minflt, rusage.ru_nswap),
            )
            loadtable.add_row(
                "Disk I/O", "%g reads, %g writes" % (rusage.ru_inblock, rusage.ru_oublock)
            )
            loadtable.add_row("Network I/O", "%g in, %g out" % (rusage.ru_msgrcv, rusage.ru_msgsnd))
            loadtable.add_row(
                "Context switching",
                "%g vol, %g forced, %g signals"
                % (rusage.ru_nvcsw, rusage.ru_nivcsw, rusage.ru_nsignals),
            )

        # os-generic

        string = "|wServer CPU and Memory load:|n\n%s" % loadtable

        # object cache count (note that sys.getsiseof is not called so this works for pypy too.
        total_num, cachedict = _IDMAPPER.cache_size()
        sorted_cache = sorted(
            [(key, num) for key, num in cachedict.items() if num > 0],
            key=lambda tup: tup[1],
            reverse=True,
        )
        memtable = self.styled_table("entity name", "number", "idmapper %", align="l")
        for tup in sorted_cache:
            memtable.add_row(tup[0], "%i" % tup[1], "%.2f" % (float(tup[1]) / total_num * 100))

        string += "\n|w Entity idmapper cache:|n %i items\n%s" % (total_num, memtable)

        # return to caller
        self.msg(string)


class CmdTickers(COMMAND_DEFAULT_CLASS):
    """
    View running tickers

    Usage:
      tickers

    Note: Tickers are created, stopped and manipulated in Python code
    using the TickerHandler. This is merely a convenience function for
    inspecting the current status.

    """

    key = "@tickers"
    help_category = "System"
    locks = "cmd:perm(tickers) or perm(Builder)"

    def func(self):
        from evennia import TICKER_HANDLER

        all_subs = TICKER_HANDLER.all_display()
        if not all_subs:
            self.msg("No tickers are currently active.")
            return
        table = self.styled_table("interval (s)", "object", "path/methodname", "idstring", "db")
        for sub in all_subs:
            table.add_row(
                sub[3],
                "%s%s"
                % (
                    sub[0] or "[None]",
                    sub[0] and " (#%s)" % (sub[0].id if hasattr(sub[0], "id") else "") or "",
                ),
                sub[1] if sub[1] else sub[2],
                sub[4] or "[Unset]",
                "*" if sub[5] else "-",
            )
        self.msg("|wActive tickers|n:\n" + str(table))


class CmdTasks(COMMAND_DEFAULT_CLASS):
    """
    Display or terminate active tasks (delays).

    Usage:
        tasks[/switch] [task_id or function_name]

    Switches:
        pause   - Pause the callback of a task.
        unpause - Process all callbacks made since pause() was called.
        do_task - Execute the task (call its callback).
        call    - Call the callback of this task.
        remove  - Remove a task without executing it.
        cancel  - Stop a task from automatically executing.

    Notes:
        A task is a single use method of delaying the call of a function. Calls are created
        in code, using `evennia.utils.delay`.
        See |luhttps://www.evennia.com/docs/latest/Command-Duration.html|ltthe docs|le for help.

        By default, tasks that are canceled and never called are cleaned up after one minute.

    Examples:
        - `tasks/cancel move_callback` - Cancels all movement delays from the slow_exit contrib.
            In this example slow exits creates it's tasks with
            `utils.delay(move_delay, move_callback)`
        - `tasks/cancel 2` - Cancel task id 2.

    """

    key = "@tasks"
    aliases = ["@delays", "@task"]
    switch_options = ("pause", "unpause", "do_task", "call", "remove", "cancel")
    locks = "perm(Developer)"
    help_category = "System"

    @staticmethod
    def coll_date_func(task):
        """Replace regex characters in date string and collect deferred function name."""
        t_comp_date = str(task[0]).replace("-", "/")
        t_func_name = str(task[1]).split(" ")
        t_func_mem_ref = t_func_name[3] if len(t_func_name) >= 4 else None
        return t_comp_date, t_func_mem_ref

    def do_task_action(self, *args, **kwargs):
        """
        Process the action of a tasks command.

        This exists to gain support with yes or no function from EvMenu.
        """
        task_id = self.task_id

        # get a reference of the global task handler
        global _TASK_HANDLER
        if _TASK_HANDLER is None:
            from evennia.scripts.taskhandler import TASK_HANDLER as _TASK_HANDLER

        # verify manipulating the correct task
        task_args = _TASK_HANDLER.tasks.get(task_id, False)
        if not task_args:  # check if the task is still active
            self.msg("Task completed while waiting for input.")
            return
        else:
            # make certain a task with matching IDs has not been created
            t_comp_date, t_func_mem_ref = self.coll_date_func(task_args)
            if self.t_comp_date != t_comp_date or self.t_func_mem_ref != t_func_mem_ref:
                self.msg("Task completed while waiting for input.")
                return

        # Do the action requested by command caller
        action_return = self.task_action()
        self.msg(f"{self.action_request} request completed.")
        self.msg(f"The task function {self.action_request} returned: {action_return}")

    def func(self):
        # get a reference of the global task handler
        global _TASK_HANDLER
        if _TASK_HANDLER is None:
            from evennia.scripts.taskhandler import TASK_HANDLER as _TASK_HANDLER
        # handle no tasks active.
        if not _TASK_HANDLER.tasks:
            self.msg("There are no active tasks.")
            if self.switches or self.args:
                self.msg("Likely the task has completed and been removed.")
            return

        # handle caller's request to manipulate a task(s)
        if self.switches and self.lhs:
            # find if the argument is a task id or function name
            action_request = self.switches[0]
            try:
                arg_is_id = int(self.lhslist[0])
            except ValueError:
                arg_is_id = False

            # if the argument is a task id, proccess the action on a single task
            if arg_is_id:
                err_arg_msg = "Switch and task ID are required when manipulating a task."
                task_comp_msg = "Task completed while processing request."

                # handle missing arguments or switches
                if not self.switches and self.lhs:
                    self.msg(err_arg_msg)
                    return

                # create a handle for the task
                task_id = arg_is_id
                task = TaskHandlerTask(task_id)

                # handle task no longer existing
                if not task.exists():
                    self.msg(f"Task {task_id} does not exist.")
                    return

                # get a reference of the function caller requested
                switch_action = getattr(task, action_request, False)
                if not switch_action:
                    self.msg(
                        f"{self.switches[0]}, is not an acceptable task action or "
                        f"{task_comp_msg.lower()}"
                    )

                # verify manipulating the correct task
                if task_id in _TASK_HANDLER.tasks:
                    task_args = _TASK_HANDLER.tasks.get(task_id, False)
                    if not task_args:  # check if the task is still active
                        self.msg(task_comp_msg)
                        return
                    else:
                        t_comp_date, t_func_mem_ref = self.coll_date_func(task_args)
                        t_func_name = str(task_args[1]).split(" ")
                        t_func_name = t_func_name[1] if len(t_func_name) >= 2 else None

                if task.exists():  # make certain the task has not been called yet.
                    prompt = (
                        f"{action_request.capitalize()} task {task_id} with completion date "
                        f"{t_comp_date} ({t_func_name}) {{options}}?"
                    )
                    no_msg = f"No {action_request} processed."
                    # record variables for use in do_task_action method
                    self.task_id = task_id
                    self.t_comp_date = t_comp_date
                    self.t_func_mem_ref = t_func_mem_ref
                    self.task_action = switch_action
                    self.action_request = action_request
                    ask_yes_no(
                        self.caller,
                        prompt=prompt,
                        yes_action=self.do_task_action,
                        no_action=no_msg,
                        default="Y",
                        allow_abort=True,
                    )
                    return True
                else:
                    self.msg(task_comp_msg)
                    return

            # the argument is not a task id, process the action on all task deferring the function
            # specified as an argument
            else:
                name_match_found = False
                arg_func_name = self.lhslist[0].lower()

                # repack tasks into a new dictionary
                current_tasks = {}
                for task_id, task_args in _TASK_HANDLER.tasks.items():
                    current_tasks.update({task_id: task_args})

                # call requested action on all tasks with the function name
                for task_id, task_args in current_tasks.items():
                    t_func_name = str(task_args[1]).split(" ")
                    t_func_name = t_func_name[1] if len(t_func_name) >= 2 else None
                    # skip this task if it is not for the function desired
                    if arg_func_name != t_func_name:
                        continue
                    name_match_found = True
                    task = TaskHandlerTask(task_id)
                    switch_action = getattr(task, action_request, False)
                    if switch_action:
                        action_return = switch_action()
                        self.msg(f"Task action {action_request} completed on task ID {task_id}.")
                        self.msg(f"The task function {action_request} returned: {action_return}")

                # provide a message if not tasks of the function name was found
                if not name_match_found:
                    self.msg(f"No tasks deferring function name {arg_func_name} found.")
                    return
                return True

        # check if an maleformed request was created
        elif self.switches or self.lhs:
            self.msg("Task command misformed.")
            self.msg("Proper format tasks[/switch] [function name or task id]")
            return

        # No task manupilation requested, build a table of tasks and display it
        # get the width of screen in characters
        width = self.client_width()
        # create table header and list to hold tasks data and actions
        tasks_header = (
            "Task ID",
            "Completion Date",
            "Function",
            "Arguments",
            "KWARGS",
            "persistent",
        )
        # empty list of lists, the size of the header
        tasks_list = [list() for i in range(len(tasks_header))]
        for task_id, task in _TASK_HANDLER.tasks.items():
            # collect data from the task
            t_comp_date, t_func_mem_ref = self.coll_date_func(task)
            t_func_name = str(task[1]).split(" ")
            t_func_name = t_func_name[1] if len(t_func_name) >= 2 else None
            t_args = str(task[2])
            t_kwargs = str(task[3])
            t_pers = str(task[4])
            # add task data to the tasks list
            task_data = (task_id, t_comp_date, t_func_name, t_args, t_kwargs, t_pers)
            for i in range(len(tasks_header)):
                tasks_list[i].append(task_data[i])
        # create and display the table
        tasks_table = EvTable(
            *tasks_header, table=tasks_list, maxwidth=width, border="cells", align="c"
        )
        actions = (f"/{switch}" for switch in self.switch_options)
        helptxt = f"\nActions: {iter_to_str(actions)}"
        self.msg(str(tasks_table) + helptxt)



"""
Link Evennia to external resources (wizard plugin for evennia_launcher)

"""
import pprint
import sys
from os import path

from django.conf import settings

from evennia.utils.utils import list_to_string, mod_import


class ConnectionWizard(object):
    def __init__(self):
        self.data = {}
        self.prev_node = None

    def display(self, text):
        "Show text"
        print(text)

    def ask_continue(self):
        "'Press return to continue'-prompt"
        input(" (Press return to continue)")

    def ask_node(self, options, prompt="Enter choice: ", default=None):
        """
        Retrieve options and jump to different menu nodes

        Args:
            options (dict): Node options on the form {key: (desc, callback), }
            prompt (str, optional): Question to ask
            default (str, optional): Default value to use if user hits return.

        """

        opt_txt = "\n".join(f" {key}: {desc}" for key, (desc, _, _) in options.items())
        self.display(opt_txt + "\n")

        while True:
            resp = input(prompt).strip()

            if not resp:
                if default:
                    resp = str(default)

            if resp.lower() in options:
                # self.display(f" Selected '{resp}'.")
                desc, callback, kwargs = options[resp.lower()]
                callback(self, **kwargs)
            elif resp.lower() in ("quit", "q"):
                sys.exit()
            elif resp:
                # input, but nothing was recognized
                self.display(" Choose one of: {}".format(list_to_string(list(options))))

    def ask_yesno(self, prompt, default="yes"):
        """
        Ask a yes/no question inline.

        Keyword Args:
            prompt (str): The prompt to ask.
            default (str): "yes" or "no", used if pressing return.
        Returns:
            reply (str): Either 'yes' or 'no'.

        """
        prompt = prompt + (" [Y]/N? " if default == "yes" else " Y/[N]? ")

        while True:
            resp = input(prompt).lstrip().lower()
            if not resp:
                resp = default.lower()
            if resp in ("yes", "y"):
                self.display(" Answered Yes.")
                return "yes"
            elif resp in ("no", "n"):
                self.display(" Answered No.")
                return "no"
            elif resp.lower() in ("quit", "q"):
                sys.exit()

    def ask_choice(self, prompt=" > ", options=None, default=None):
        """
        Ask multiple-choice question, get response inline.

        Keyword Args:
            prompt (str): Input prompt.
            options (list): List of options. Will be indexable by sequence number 1...
            default (int): The list index+1 of the default choice, if any
        Returns:
            reply (str): The answered reply.

        """
        opt_txt = "\n".join(f" {ind + 1}: {desc}" for ind, desc in enumerate(options))
        self.display(opt_txt + "\n")

        while True:
            resp = input(prompt).strip()

            if not resp:
                if default:
                    return options[int(default)]
            if resp.lower() in ("quit", "q"):
                sys.exit()
            if resp.isdigit():
                resp = int(resp) - 1
                if 0 <= resp < len(options):
                    selection = options[resp]
                    self.display(f" Selected '{selection}'.")
                    return selection
            self.display(" Select one of the given options.")

    def ask_input(self, prompt=" > ", default=None, validator=None):
        """
        Get arbitrary input inline.

        Keyword Args:
            prompt (str): The display prompt.
            default (str): If empty input, use this.
            validator (callable): If given, the input will be passed
                into this callable. It should return True unless validation
                fails (and is expected to echo why if so).

        Returns:
            inp (str): The input given, or default.

        """
        while True:
            resp = input(prompt).strip()

            if not resp and default:
                resp = str(default)

            if resp.lower() in ("q", "quit"):
                sys.exit()

            if resp.lower() == "none":
                resp = ""

                if validator and not validator(resp):
                    continue

                ok = input("\n Leave blank? [Y]/N: ")
                if ok.lower() in ("n", "no"):
                    continue
                elif ok.lower() in ("q", "quit"):
                    sys.exit()
                return resp

            if validator and not validator(resp):
                continue

            self.display(resp)
            ok = input("\n Is this correct? [Y]/N: ")
            if ok.lower() in ("n", "no"):
                continue
            elif ok.lower() in ("q", "quit"):
                sys.exit()
            return resp


def node_start(wizard):
    text = """
    This wizard helps to attach your Evennia server to external networks. It
    will save to a file `server/conf/connection_settings.py` that will be
    imported from the bottom of your game settings file. Once generated you can
    also modify that file directly.

    Make sure you have at least started the game once before continuing!

    Use `quit` at any time to abort and throw away unsaved changes.
    """
    options = {
        "1": (
            "Add your game to the Evennia game index (also for closed-dev games)",
            node_game_index_start,
            {},
        ),
        "2": ("MSSP setup (for mud-list crawlers)", node_mssp_start, {}),
        # "3": ("Add Grapevine listing",
        #       node_grapevine_start, {}),
        # "4": ("Add IRC link",
        #      "node_irc_start", {}),
        # "5" ("Add RSS feed",
        #       "node_rss_start", {}),
        "s": ("View and (optionally) Save created settings", node_view_and_apply_settings, {}),
        "q": ("Quit", lambda *args: sys.exit(), {}),
    }

    wizard.display(text)
    wizard.ask_node(options)


# Evennia game index


def node_game_index_start(wizard, **kwargs):
    text = """
    The Evennia game index (http://games.evennia.com) lists both active Evennia
    games as well as games in various stages of development.

    You can put up your game in the index also if you are not (yet) open for
    players.  If so, put 'None' for the connection details - you are just telling
    us that you are out there, making us excited about your upcoming game!

    Please check the listing online first to see that your exact game name is
    not colliding with an existing game-name in the list (be nice!).
    """

    wizard.display(text)
    if wizard.ask_yesno("Continue adding/editing an Index entry?") == "yes":
        node_game_index_fields(wizard)
    else:
        node_start(wizard)


def node_game_index_fields(wizard, status=None):
    # reset the listing if needed
    if not hasattr(wizard, "game_index_listing"):
        wizard.game_index_listing = settings.GAME_INDEX_LISTING

    # game status

    status_default = wizard.game_index_listing["game_status"]
    text = f"""
    What is the status of your game?
        - pre-alpha: a game in its very early stages, mostly unfinished or unstarted
        - alpha: a working concept, probably lots of bugs and incomplete features
        - beta: a working game, but expect bugs and changing features
        - launched: a full, working game (that may still be expanded upon and improved later)

    Current value (return to keep):
    {status_default}
    """

    options = ["pre-alpha", "alpha", "beta", "launched"]

    wizard.display(text)
    wizard.game_index_listing["game_status"] = wizard.ask_choice("Select one: ", options)

    # game name

    name_default = settings.SERVERNAME
    text = f"""
    Your game's name should usually be the same as `settings.SERVERNAME`, but
    you can set it to something else here if you want.

    Current value:
    {name_default}
    """

    def name_validator(inp):
        tmax = 80
        tlen = len(inp)
        if tlen > tmax:
            print(f"The name must be shorter than {tmax} characters (was {tlen}).")
            wizard.ask_continue()
            return False
        return True

    wizard.display(text)
    wizard.game_index_listing["game_name"] = wizard.ask_input(
        default=name_default, validator=name_validator
    )

    # short desc

    sdesc_default = wizard.game_index_listing.get("short_description", None)

    text = f"""
    Enter a short description of your game. Make it snappy and interesting!
    This should be at most one or two sentences (255 characters) to display by
    '{settings.SERVERNAME}' in the main game list. Line breaks will be ignored.

    Current value:
    {sdesc_default}
    """

    def sdesc_validator(inp):
        tmax = 255
        tlen = len(inp)
        if tlen > tmax:
            print(f"The short desc must be shorter than {tmax} characters (was {tlen}).")
            wizard.ask_continue()
            return False
        return True

    wizard.display(text)
    wizard.game_index_listing["short_description"] = wizard.ask_input(
        default=sdesc_default, validator=sdesc_validator
    )

    # long desc

    long_default = wizard.game_index_listing.get("long_description", None)

    text = f"""
    Enter a longer, full-length description. This will be shown when clicking
    on your game's listing. You can use \\n to create line breaks and may use
    Markdown formatting like *bold*, _italic_, [linkname](http://link) etc.

    Current value:
    {long_default}
    """

    wizard.display(text)
    wizard.game_index_listing["long_description"] = wizard.ask_input(default=long_default)

    # listing contact

    listing_default = wizard.game_index_listing.get("listing_contact", None)
    text = f"""
    Enter a listing email-contact. This will not be visible in the listing, but
    allows us to get in touch with you should there be some listing issue (like
    a name collision) or some bug with the listing (us actually using this is
    likely to be somewhere between super-rarely and never).

    Current value:
    {listing_default}
    """

    def contact_validator(inp):
        if not inp or "@" not in inp:
            print("This should be an email and cannot be blank.")
            wizard.ask_continue()
            return False
        return True

    wizard.display(text)
    wizard.game_index_listing["listing_contact"] = wizard.ask_input(
        default=listing_default, validator=contact_validator
    )

    # telnet hostname

    hostname_default = wizard.game_index_listing.get("telnet_hostname", None)
    text = f"""
    Enter the hostname to which third-party telnet mud clients can connect to
    your game. This would be the name of the server your game is hosted on,
    like `coolgame.games.com`, or `mygreatgame.se`.

    Write 'None' if you are not offering public telnet connections at this time.

    Current value:
    {hostname_default}
    """

    wizard.display(text)
    wizard.game_index_listing["telnet_hostname"] = wizard.ask_input(default=hostname_default)

    # telnet port

    port_default = wizard.game_index_listing.get("telnet_port", None)
    text = f"""
    Enter the main telnet port. The Evennia default is 4000. You can change
    this with the TELNET_PORTS server setting.

    Write 'None' if you are not offering public telnet connections at this time.

    Current value:
    {port_default}
    """

    wizard.display(text)
    wizard.game_index_listing["telnet_port"] = wizard.ask_input(default=port_default)

    # website

    website_default = wizard.game_index_listing.get("game_website", None)
    text = f"""
    Evennia is its own web server and runs your game's website. Enter the
    URL of the website here, like http://yourwebsite.com, here.

    Write 'None' if you are not offering a publicly visible website at this time.

    Current value:
    {website_default}
    """

    wizard.display(text)
    wizard.game_index_listing["game_website"] = wizard.ask_input(default=website_default)

    # webclient

    webclient_default = wizard.game_index_listing.get("web_client_url", None)
    text = f"""
    Evennia offers its own native webclient. Normally it will be found from the
    game homepage at something like http://yourwebsite.com/webclient. Enter
    your specific URL here (when clicking this link you should launch into the
    web client)

    Write 'None' if you don't want to list a publicly accessible webclient.

    Current value:
    {webclient_default}
    """

    wizard.display(text)
    wizard.game_index_listing["web_client_url"] = wizard.ask_input(default=webclient_default)

    if not (
        wizard.game_index_listing.get("web_client_url")
        or (wizard.game_index_listing.get("telnet_host"))
    ):
        wizard.display(
            "\nNote: You have not specified any connection options. This means "
            "your game \nwill be marked as being in 'closed development' in "
            "the index."
        )

    wizard.display("\nDon't forget to inspect and save your changes.")

    node_start(wizard)


# MSSP


def node_mssp_start(wizard):
    mssp_module = mod_import(settings.MSSP_META_MODULE or "server.conf.mssp")
    try:
        filename = mssp_module.__file__
    except AttributeError:
        filename = "server/conf/mssp.py"

    text = f"""
    MSSP (Mud Server Status Protocol) has a vast amount of options so it must
    be modified outside this wizard by directly editing its config file here:

        '{filename}'

    MSSP allows traditional online MUD-listing sites/crawlers to continuously
    monitor your game and list information about it. Some of this, like active
    player-count, Evennia will automatically add for you, whereas most fields
    you need to set manually.

    To use MSSP you should generally have a publicly open game that external
    players can connect to. You also need to register at a MUD listing site to
    tell them to crawl your game.
    """

    wizard.display(text)
    wizard.ask_continue()
    node_start(wizard)


# Admin


def _save_changes(wizard):
    """
    Perform the save
    """

    # add import statement to settings file
    import_stanza = "from .connection_settings import *"
    setting_module = mod_import("server.conf.settings")
    with open(setting_module.__file__, "r+") as f:
        txt = f.read()  # moves pointer to end of file
        if import_stanza not in txt:
            # add to the end of the file
            f.write(
                "\n\n"
                "try:\n"
                "    # Created by the `evennia connections` wizard\n"
                f"    {import_stanza}\n"
                "except ImportError:\n"
                "    pass"
            )

    connect_settings_file = path.join(settings.GAME_DIR, "server", "conf", "connection_settings.py")
    with open(connect_settings_file, "w") as f:
        f.write(
            "# This file is auto-generated by the `evennia connections` wizard.\n"
            "# Don't edit manually, your changes will be overwritten.\n\n"
        )

        f.write(wizard.save_output)
        wizard.display(f"saving to {connect_settings_file} ...")


def node_view_and_apply_settings(wizard):
    """
    Inspect and save the data gathered in the other nodes

    """
    pp = pprint.PrettyPrinter(indent=4)
    saves = False

    # game index
    game_index_save_text = ""
    game_index_listing = (
        wizard.game_index_listing if hasattr(wizard, "game_index_listing") else None
    )
    if not game_index_listing and settings.GAME_INDEX_ENABLED:
        game_index_listing = settings.GAME_INDEX_LISTING
    if game_index_listing:
        game_index_save_text = (
            "GAME_INDEX_ENABLED = True\n"
            "GAME_INDEX_LISTING = \\\n" + pp.pformat(game_index_listing)
        )
        saves = True
    else:
        game_index_save_text = "# No Game Index settings found."

    # potentially add other wizards in the future
    text = game_index_save_text

    wizard.display(f"Settings to save:\n\n{text}")

    if saves:
        if wizard.ask_yesno("\nDo you want to save these settings?") == "yes":
            wizard.save_output = text
            _save_changes(wizard)
            wizard.display("... saved!\nThe changes will apply after you reload your server.")
        else:
            wizard.display("... cancelled.")
    wizard.ask_continue()
    node_start(wizard)



"""
The client for sending data to the Evennia Game Index

"""
import platform
import urllib.error
import urllib.parse
import urllib.request

import django
from django.conf import settings
from twisted.internet import defer, protocol, reactor
from twisted.internet.defer import inlineCallbacks
from twisted.web.client import Agent, HTTPConnectionPool, _HTTP11ClientFactory
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from zope.interface import implementer

import evennia
from evennia.accounts.models import AccountDB
from evennia.utils import get_evennia_version, logger

_EGI_HOST = "http://evennia-game-index.appspot.com"
_EGI_REPORT_PATH = "/api/v1/game/check_in"


class EvenniaGameIndexClient:
    """
    This client class is used for gathering and sending game details to the
    Evennia Game Index. Since EGI is in the early goings, this isn't
    incredibly configurable as far as to what is being sent.
    """

    def __init__(self, on_bad_request=None):
        """
        on_bad_request (callable, optional): Callable to trigger when a bad request was sent.

        """
        self.report_host = _EGI_HOST
        self.report_path = _EGI_REPORT_PATH
        self.report_url = self.report_host + self.report_path
        self.logged_first_connect = False

        self._on_bad_request = on_bad_request
        # Oh, the humanity. Silence the factory start/stop messages.
        self._conn_pool = HTTPConnectionPool(reactor)
        self._conn_pool._factory = QuietHTTP11ClientFactory

    @inlineCallbacks
    def send_game_details(self):
        """
        This is where the magic happens. Send details about the game to the
        Evennia Game Index.
        """
        status_code, response_body = yield self._form_and_send_request()
        if status_code == 200:
            if not self.logged_first_connect:
                logger.log_infomsg("Successfully sent game details to Evennia Game Index.")
                self.logged_first_connect = True
            return
        # At this point, either EGD is having issues or the payload we sent
        # is improperly formed (probably due to mis-configuration).
        logger.log_errmsg(
            "Failed to send game details to Evennia Game Index. HTTP "
            "status code was %s. Message was: %s" % (status_code, response_body)
        )

        if status_code == 400 and self._on_bad_request:
            # Improperly formed request. Defer to the callback as far as what
            # to do. Probably not a great idea to continue attempting to send
            # to EGD, though.
            self._on_bad_request()

    def _form_and_send_request(self):
        """
        Build the request to send to the index.

        """
        agent = Agent(reactor, pool=self._conn_pool)
        headers = {
            b"User-Agent": [b"Evennia Game Index Client"],
            b"Content-Type": [b"application/x-www-form-urlencoded"],
        }
        egi_config = settings.GAME_INDEX_LISTING
        # We are using `or` statements below with dict.get() to avoid sending
        # stringified 'None' values to the server.
        try:
            values = {
                # Game listing stuff
                "game_name": egi_config.get("game_name", settings.SERVERNAME),
                "game_status": egi_config["game_status"],
                "game_website": egi_config.get("game_website", ""),
                "short_description": egi_config["short_description"],
                "long_description": egi_config.get("long_description", ""),
                "listing_contact": egi_config["listing_contact"],
                # How to play
                "telnet_hostname": egi_config.get("telnet_hostname", ""),
                "telnet_port": egi_config.get("telnet_port", ""),
                "web_client_url": egi_config.get("web_client_url", ""),
                # Game stats
                "connected_account_count": evennia.SESSION_HANDLER.account_count(),
                "total_account_count": AccountDB.objects.num_total_accounts() or 0,
                # System info
                "evennia_version": get_evennia_version(),
                "python_version": platform.python_version(),
                "django_version": django.get_version(),
                "server_platform": platform.platform(),
            }
        except KeyError as err:
            raise KeyError(f"Error loading GAME_INDEX_LISTING: {err}")

        data = urllib.parse.urlencode(values)

        d = agent.request(
            b"POST",
            bytes(self.report_url, "utf-8"),
            headers=Headers(headers),
            bodyProducer=StringProducer(data),
        )

        d.addCallback(self.handle_egd_response)
        return d

    def handle_egd_response(self, response):
        if 200 <= response.code < 300:
            d = defer.succeed((response.code, "OK"))
        else:
            # Go through the horrifying process of getting the response body
            # out of Twisted's plumbing.
            d = defer.Deferred()
            response.deliverBody(SimpleResponseReceiver(response.code, d))
        return d


class SimpleResponseReceiver(protocol.Protocol):
    """
    Used for pulling the response body out of an HTTP response.
    """

    def __init__(self, status_code, d):
        self.status_code = status_code
        self.buf = ""
        self.d = d

    def dataReceived(self, data):
        self.buf += data

    def connectionLost(self, reason=protocol.connectionDone):
        self.d.callback((self.status_code, self.buf))


@implementer(IBodyProducer)
class StringProducer:
    """
    Used for feeding a request body to the tx HTTP client.
    """

    def __init__(self, body):
        self.body = bytes(body, "utf-8")
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return defer.succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class QuietHTTP11ClientFactory(_HTTP11ClientFactory):
    """
    Silences the obnoxious factory start/stop messages in the default client.
    """

    noisy = False



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("objects", "0002_auto_20140917_0756")]

    operations = [
        migrations.CreateModel(
            name="DefaultObject", fields=[], options={"proxy": True}, bases=("objects.objectdb",)
        ),
        migrations.CreateModel(
            name="DefaultExit", fields=[], options={"proxy": True}, bases=("objects.defaultobject",)
        ),
        migrations.CreateModel(
            name="DefaultCharacter",
            fields=[],
            options={"proxy": True},
            bases=("objects.defaultobject",),
        ),
        migrations.CreateModel(
            name="DefaultRoom", fields=[], options={"proxy": True}, bases=("objects.defaultobject",)
        ),
    ]
